# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

from ..aaz.latest.monitor.diagnostic_settings import Create as _DiagnosticSettingsCreate
from ..aaz.latest.monitor.diagnostic_settings import List as _DiagnosticSettingsList
from ..aaz.latest.monitor.diagnostic_settings import Show as _DiagnosticSettingsShow
from ..aaz.latest.monitor.diagnostic_settings import Delete as _DiagnosticSettingsDelete
from ..aaz.latest.monitor.diagnostic_settings import Update as _DiagnosticSettingsUpdate
from ..aaz.latest.monitor.diagnostic_settings.categories import List as _DiagnosticSettingsCategoryList
from ..aaz.latest.monitor.diagnostic_settings.categories import Show as _DiagnosticSettingsCategoryShow
from knack.util import CLIError
from azure.cli.core.azclierror import ArgumentUsageError


def create_resource_parameters(arg_schema, arg_group=None):
    from azure.cli.core.aaz import AAZResourceGroupNameArg, AAZStrArg
    arg_schema.resource_group_name = AAZResourceGroupNameArg(arg_group=arg_group)
    arg_schema.namespace = AAZStrArg(
        options=['--resource-namespace'],
        help="Target resource provider namespace.",
        arg_group=arg_group,
    )
    arg_schema.parent = AAZStrArg(
        options=['--resource-parent'],
        help="Target resource parent path, if applicable.",
        arg_group=arg_group,
    )
    arg_schema.resource_type = AAZStrArg(
        options=['--resource-type'],
        help="Target resource type. Can also accept namespace/type format (Ex: 'Microsoft.Compute/virtualMachines')",
        arg_group=arg_group,
    )


def update_resource_parameters(ctx, alias="resource"):
    args = ctx.args
    from azure.mgmt.core.tools import parse_resource_id, is_valid_resource_id
    from azure.cli.core.aaz import has_value
    name_or_id = args.resource.to_serialized_data()
    usage_error = CLIError('usage error: --{0} ID | --{0} NAME --resource-group NAME '
                           '--{0}-type TYPE [--{0}-parent PARENT] '
                           '[--{0}-namespace NAMESPACE]'.format(alias))
    if not name_or_id:
        raise usage_error
    if is_valid_resource_id(name_or_id):
        if has_value(args.namespace) or has_value(args.parent) or has_value(args.resource_type):
            raise usage_error
        parsed_id = parse_resource_id(name_or_id)
        subscription_id = parsed_id.get('subscription', None)
        if subscription_id:
            # update subscription_id to support cross tenants
            ctx.update_aux_subscriptions(subscription_id)
    else:
        res_ns = args.namespace.to_serialized_data()
        rg = args.resource_group_name.to_serialized_data()
        res_type = args.resource_type.to_serialized_data()
        parent = args.parent.to_serialized_data()
        if has_value(res_type):
            if '/' in res_type:
                res_ns = res_ns or res_type.rsplit('/', 1)[0]
                res_type = res_type.rsplit('/', 1)[1]

        if not res_ns or not rg or not res_type or not name_or_id:
            raise usage_error

        args.resource = f"/subscriptions/{ctx.subscription_id}/resourceGroups/{rg}/providers/{res_ns}/" \
                        f"{parent + '/' if parent else ''}{res_type}/{name_or_id}"


class DiagnosticSettingsCreate(_DiagnosticSettingsCreate):
    @classmethod
    def _build_arguments_schema(cls, *args, **kwargs):
        arg_schema = super()._build_arguments_schema(*args, **kwargs)
        create_resource_parameters(arg_schema, arg_group="Target Resource")

        from azure.cli.core.aaz import AAZBoolArg
        arg_schema.export_to_resource_specific = AAZBoolArg(
            options=['--export-to-resource-specific'],
            help="Indicate that the export to LA must be done to a resource specific table, a.k.a. "
                 "dedicated or fixed schema table, as opposed to the default dynamic schema table called "
                 "AzureDiagnostics. This argument is effective only when the argument --workspace is also given."
                 "  Allowed values: false, true."
        )
        arg_schema.log_analytics_destination_type._registered = False  # pylint:disable=protected-access
        arg_schema.service_bus_rule_id._registered = False  # pylint:disable=protected-access
        return arg_schema

    def pre_operations(self):
        ctx = self.ctx
        from azure.cli.core.aaz import has_value
        from azure.mgmt.core.tools import is_valid_resource_id, resource_id, parse_resource_id
        update_resource_parameters(ctx)
        args = ctx.args
        rg = args.resource_group_name.to_serialized_data()

        if not has_value(rg):
            rg = parse_resource_id(args.resource.to_serialized_data())['resource_group']
            args.resource_group_name = rg

        storage_account = args.storage_account.to_serialized_data()
        if has_value(storage_account) and not is_valid_resource_id(storage_account):
            storage_account = resource_id(subscription=ctx.subscription_id,
                                          resource_group=rg,
                                          namespace='microsoft.Storage',
                                          type='storageAccounts',
                                          name=storage_account)
            args.storage_account = storage_account

        workspace = args.workspace.to_serialized_data()
        if has_value(workspace) and not is_valid_resource_id(workspace):
            workspace = resource_id(subscription=ctx.subscription_id,
                                    resource_group=rg,
                                    namespace='microsoft.OperationalInsights',
                                    type='workspaces',
                                    name=workspace)
            args.workspace = workspace

        event_hub = args.event_hub.to_serialized_data()
        if has_value(event_hub) and is_valid_resource_id(event_hub):
            event_hub = parse_resource_id(event_hub)['name']
            args.event_hub = event_hub

        event_hub_rule = args.event_hub_rule.to_serialized_data()
        if has_value(event_hub_rule):
            if not is_valid_resource_id(event_hub_rule):
                if not has_value(event_hub):
                    raise CLIError('usage error: --event-hub-rule ID | --event-hub-rule NAME --event-hub NAME')
                # use value from --event-hub if the rule is a name
                event_hub_rule = resource_id(
                    subscription=ctx.subscription_id,
                    resource_group=rg,
                    namespace='Microsoft.EventHub',
                    type='namespaces',
                    name=event_hub,
                    child_type_1='AuthorizationRules',
                    child_name_1=event_hub_rule)
                args.event_hub_rule = event_hub_rule

            elif not has_value(event_hub):
                # extract the event hub name from `--event-hub-rule` if provided as an ID
                event_hub = parse_resource_id(event_hub_rule)['name']
                args.event_hub = event_hub
        if not (has_value(storage_account) or has_value(workspace) or has_value(event_hub)):
            raise CLIError(
                'usage error - expected one or more:  --storage-account NAME_OR_ID | --workspace NAME_OR_ID '
                '| --event-hub NAME_OR_ID | --event-hub-rule ID')

        export_to_resource_specific = args.export_to_resource_specific.to_serialized_data()
        if has_value(export_to_resource_specific) and export_to_resource_specific:
            args.log_analytics_destination_type = 'Dedicated'
            if not has_value(workspace):
                raise ArgumentUsageError('usage error: --workspace and --export-to-specific-resource')
        else:
            args.log_analytics_destination_type = None


class DiagnosticSettingsShow(_DiagnosticSettingsShow):

    @classmethod
    def _build_arguments_schema(cls, *args, **kwargs):
        arg_schema = super()._build_arguments_schema(*args, **kwargs)
        create_resource_parameters(arg_schema, arg_group="Target Resource")
        return arg_schema

    def pre_operations(self):
        ctx = self.ctx
        update_resource_parameters(ctx)


class DiagnosticSettingsList(_DiagnosticSettingsList):

    @classmethod
    def _build_arguments_schema(cls, *args, **kwargs):
        arg_schema = super()._build_arguments_schema(*args, **kwargs)
        create_resource_parameters(arg_schema)
        return arg_schema

    def pre_operations(self):
        ctx = self.ctx
        update_resource_parameters(ctx)


class DiagnosticSettingsDelete(_DiagnosticSettingsDelete):

    @classmethod
    def _build_arguments_schema(cls, *args, **kwargs):
        arg_schema = super()._build_arguments_schema(*args, **kwargs)
        create_resource_parameters(arg_schema, arg_group="Target Resource")
        return arg_schema

    def pre_operations(self):
        ctx = self.ctx
        update_resource_parameters(ctx)


class DiagnosticSettingsUpdate(_DiagnosticSettingsUpdate):

    @classmethod
    def _build_arguments_schema(cls, *args, **kwargs):
        arg_schema = super()._build_arguments_schema(*args, **kwargs)
        create_resource_parameters(arg_schema, arg_group="Target Resource")
        return arg_schema

    def pre_operations(self):
        ctx = self.ctx
        update_resource_parameters(ctx)


class DiagnosticSettingsCategoryList(_DiagnosticSettingsCategoryList):
    @classmethod
    def _build_arguments_schema(cls, *args, **kwargs):
        arg_schema = super()._build_arguments_schema(*args, **kwargs)
        create_resource_parameters(arg_schema)
        return arg_schema

    def pre_operations(self):
        ctx = self.ctx
        update_resource_parameters(ctx)


class DiagnosticSettingsCategoryShow(_DiagnosticSettingsCategoryShow):
    @classmethod
    def _build_arguments_schema(cls, *args, **kwargs):
        arg_schema = super()._build_arguments_schema(*args, **kwargs)
        create_resource_parameters(arg_schema)
        return arg_schema

    def pre_operations(self):
        ctx = self.ctx
        update_resource_parameters(ctx)


# pylint: disable=unused-argument, line-too-long
def create_diagnostics_settings(client, name, resource_uri,
                                logs=None,
                                metrics=None,
                                event_hub=None,
                                event_hub_rule=None,
                                storage_account=None,
                                workspace=None,
                                export_to_resource_specific=None):
    from azure.mgmt.monitor.models import DiagnosticSettingsResource
    if export_to_resource_specific and workspace is None:
        raise CLIError('usage error: --workspace and --export-to-specific-resource')
    parameters = DiagnosticSettingsResource(storage_account_id=storage_account,
                                            workspace_id=workspace,
                                            event_hub_name=event_hub,
                                            event_hub_authorization_rule_id=event_hub_rule,
                                            metrics=metrics,
                                            logs=logs,
                                            log_analytics_destination_type='Dedicated' if export_to_resource_specific else None)

    return client.create_or_update(resource_uri=resource_uri, parameters=parameters, name=name)



import functools
import sys
from typing import Callable, Generator, Iterable, Iterator, Optional, Tuple, TypeVar

from pip._vendor.rich.progress import (
    BarColumn,
    DownloadColumn,
    FileSizeColumn,
    MofNCompleteColumn,
    Progress,
    ProgressColumn,
    SpinnerColumn,
    TextColumn,
    TimeElapsedColumn,
    TimeRemainingColumn,
    TransferSpeedColumn,
)

from pip._internal.cli.spinners import RateLimiter
from pip._internal.req.req_install import InstallRequirement
from pip._internal.utils.logging import get_console, get_indentation

T = TypeVar("T")
ProgressRenderer = Callable[[Iterable[T]], Iterator[T]]


def _rich_download_progress_bar(
    iterable: Iterable[bytes],
    *,
    bar_type: str,
    size: Optional[int],
    initial_progress: Optional[int] = None,
) -> Generator[bytes, None, None]:
    assert bar_type == "on", "This should only be used in the default mode."

    if not size:
        total = float("inf")
        columns: Tuple[ProgressColumn, ...] = (
            TextColumn("[progress.description]{task.description}"),
            SpinnerColumn("line", speed=1.5),
            FileSizeColumn(),
            TransferSpeedColumn(),
            TimeElapsedColumn(),
        )
    else:
        total = size
        columns = (
            TextColumn("[progress.description]{task.description}"),
            BarColumn(),
            DownloadColumn(),
            TransferSpeedColumn(),
            TextColumn("eta"),
            TimeRemainingColumn(),
        )

    progress = Progress(*columns, refresh_per_second=5)
    task_id = progress.add_task(" " * (get_indentation() + 2), total=total)
    if initial_progress is not None:
        progress.update(task_id, advance=initial_progress)
    with progress:
        for chunk in iterable:
            yield chunk
            progress.update(task_id, advance=len(chunk))


def _rich_install_progress_bar(
    iterable: Iterable[InstallRequirement], *, total: int
) -> Iterator[InstallRequirement]:
    columns = (
        TextColumn("{task.fields[indent]}"),
        BarColumn(),
        MofNCompleteColumn(),
        TextColumn("{task.description}"),
    )
    console = get_console()

    bar = Progress(*columns, refresh_per_second=6, console=console, transient=True)
    # Hiding the progress bar at initialization forces a refresh cycle to occur
    # until the bar appears, avoiding very short flashes.
    task = bar.add_task("", total=total, indent=" " * get_indentation(), visible=False)
    with bar:
        for req in iterable:
            bar.update(task, description=rf"\[{req.name}]", visible=True)
            yield req
            bar.advance(task)


def _raw_progress_bar(
    iterable: Iterable[bytes],
    *,
    size: Optional[int],
    initial_progress: Optional[int] = None,
) -> Generator[bytes, None, None]:
    def write_progress(current: int, total: int) -> None:
        sys.stdout.write(f"Progress {current} of {total}\n")
        sys.stdout.flush()

    current = initial_progress or 0
    total = size or 0
    rate_limiter = RateLimiter(0.25)

    write_progress(current, total)
    for chunk in iterable:
        current += len(chunk)
        if rate_limiter.ready() or current == total:
            write_progress(current, total)
            rate_limiter.reset()
        yield chunk


def get_download_progress_renderer(
    *, bar_type: str, size: Optional[int] = None, initial_progress: Optional[int] = None
) -> ProgressRenderer[bytes]:
    """Get an object that can be used to render the download progress.

    Returns a callable, that takes an iterable to "wrap".
    """
    if bar_type == "on":
        return functools.partial(
            _rich_download_progress_bar,
            bar_type=bar_type,
            size=size,
            initial_progress=initial_progress,
        )
    elif bar_type == "raw":
        return functools.partial(
            _raw_progress_bar,
            size=size,
            initial_progress=initial_progress,
        )
    else:
        return iter  # no-op, when passed an iterator


def get_install_progress_renderer(
    *, bar_type: str, total: int
) -> ProgressRenderer[InstallRequirement]:
    """Get an object that can be used to render the install progress.
    Returns a callable, that takes an iterable to "wrap".
    """
    if bar_type == "on":
        return functools.partial(_rich_install_progress_bar, total=total)
    else:
        return iter



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
#
# Code generated by aaz-dev-tools
# --------------------------------------------------------------------------------------------

# pylint: skip-file
# flake8: noqa

from azure.cli.core.aaz import *


@register_command(
    "disk-encryption-set wait",
)
class Wait(AAZWaitCommand):
    """Place the CLI in a waiting state until a condition is met.
    """

    _aaz_info = {
        "resources": [
            ["mgmt-plane", "/subscriptions/{}/resourcegroups/{}/providers/microsoft.compute/diskencryptionsets/{}", "2023-04-02"],
        ]
    }

    def _handler(self, command_args):
        super()._handler(command_args)
        self._execute_operations()
        return self._output()

    _args_schema = None

    @classmethod
    def _build_arguments_schema(cls, *args, **kwargs):
        if cls._args_schema is not None:
            return cls._args_schema
        cls._args_schema = super()._build_arguments_schema(*args, **kwargs)

        # define Arg Group ""

        _args_schema = cls._args_schema
        _args_schema.disk_encryption_set_name = AAZStrArg(
            options=["-n", "--name", "--disk-encryption-set-name"],
            help="Name of disk encryption set.",
            required=True,
            id_part="name",
        )
        _args_schema.resource_group = AAZResourceGroupNameArg(
            required=True,
        )
        return cls._args_schema

    def _execute_operations(self):
        self.pre_operations()
        self.DiskEncryptionSetsGet(ctx=self.ctx)()
        self.post_operations()

    @register_callback
    def pre_operations(self):
        pass

    @register_callback
    def post_operations(self):
        pass

    def _output(self, *args, **kwargs):
        result = self.deserialize_output(self.ctx.vars.instance, client_flatten=False)
        return result

    class DiskEncryptionSetsGet(AAZHttpOperation):
        CLIENT_TYPE = "MgmtClient"

        def __call__(self, *args, **kwargs):
            request = self.make_request()
            session = self.client.send_request(request=request, stream=False, **kwargs)
            if session.http_response.status_code in [200]:
                return self.on_200(session)

            return self.on_error(session.http_response)

        @property
        def url(self):
            return self.client.format_url(
                "/subscriptions/{subscriptionId}/resourceGroups/{resourceGroupName}/providers/Microsoft.Compute/diskEncryptionSets/{diskEncryptionSetName}",
                **self.url_parameters
            )

        @property
        def method(self):
            return "GET"

        @property
        def error_format(self):
            return "ODataV4Format"

        @property
        def url_parameters(self):
            parameters = {
                **self.serialize_url_param(
                    "diskEncryptionSetName", self.ctx.args.disk_encryption_set_name,
                    required=True,
                ),
                **self.serialize_url_param(
                    "resourceGroupName", self.ctx.args.resource_group,
                    required=True,
                ),
                **self.serialize_url_param(
                    "subscriptionId", self.ctx.subscription_id,
                    required=True,
                ),
            }
            return parameters

        @property
        def query_parameters(self):
            parameters = {
                **self.serialize_query_param(
                    "api-version", "2023-04-02",
                    required=True,
                ),
            }
            return parameters

        @property
        def header_parameters(self):
            parameters = {
                **self.serialize_header_param(
                    "Accept", "application/json",
                ),
            }
            return parameters

        def on_200(self, session):
            data = self.deserialize_http_content(session)
            self.ctx.set_var(
                "instance",
                data,
                schema_builder=self._build_schema_on_200
            )

        _schema_on_200 = None

        @classmethod
        def _build_schema_on_200(cls):
            if cls._schema_on_200 is not None:
                return cls._schema_on_200

            cls._schema_on_200 = AAZObjectType()

            _schema_on_200 = cls._schema_on_200
            _schema_on_200.id = AAZStrType(
                flags={"read_only": True},
            )
            _schema_on_200.identity = AAZIdentityObjectType()
            _schema_on_200.location = AAZStrType(
                flags={"required": True},
            )
            _schema_on_200.name = AAZStrType(
                flags={"read_only": True},
            )
            _schema_on_200.properties = AAZObjectType(
                flags={"client_flatten": True},
            )
            _schema_on_200.tags = AAZDictType()
            _schema_on_200.type = AAZStrType(
                flags={"read_only": True},
            )

            identity = cls._schema_on_200.identity
            identity.principal_id = AAZStrType(
                serialized_name="principalId",
                flags={"read_only": True},
            )
            identity.tenant_id = AAZStrType(
                serialized_name="tenantId",
                flags={"read_only": True},
            )
            identity.type = AAZStrType()
            identity.user_assigned_identities = AAZDictType(
                serialized_name="userAssignedIdentities",
            )

            user_assigned_identities = cls._schema_on_200.identity.user_assigned_identities
            user_assigned_identities.Element = AAZObjectType()

            _element = cls._schema_on_200.identity.user_assigned_identities.Element
            _element.client_id = AAZStrType(
                serialized_name="clientId",
                flags={"read_only": True},
            )
            _element.principal_id = AAZStrType(
                serialized_name="principalId",
                flags={"read_only": True},
            )

            properties = cls._schema_on_200.properties
            properties.active_key = AAZObjectType(
                serialized_name="activeKey",
            )
            _WaitHelper._build_schema_key_for_disk_encryption_set_read(properties.active_key)
            properties.auto_key_rotation_error = AAZObjectType(
                serialized_name="autoKeyRotationError",
                flags={"read_only": True},
            )
            _WaitHelper._build_schema_api_error_read(properties.auto_key_rotation_error)
            properties.encryption_type = AAZStrType(
                serialized_name="encryptionType",
            )
            properties.federated_client_id = AAZStrType(
                serialized_name="federatedClientId",
            )
            properties.last_key_rotation_timestamp = AAZStrType(
                serialized_name="lastKeyRotationTimestamp",
                flags={"read_only": True},
            )
            properties.previous_keys = AAZListType(
                serialized_name="previousKeys",
                flags={"read_only": True},
            )
            properties.provisioning_state = AAZStrType(
                serialized_name="provisioningState",
                flags={"read_only": True},
            )
            properties.rotation_to_latest_key_version_enabled = AAZBoolType(
                serialized_name="rotationToLatestKeyVersionEnabled",
            )

            previous_keys = cls._schema_on_200.properties.previous_keys
            previous_keys.Element = AAZObjectType()
            _WaitHelper._build_schema_key_for_disk_encryption_set_read(previous_keys.Element)

            tags = cls._schema_on_200.tags
            tags.Element = AAZStrType()

            return cls._schema_on_200


class _WaitHelper:
    """Helper class for Wait"""

    _schema_api_error_read = None

    @classmethod
    def _build_schema_api_error_read(cls, _schema):
        if cls._schema_api_error_read is not None:
            _schema.code = cls._schema_api_error_read.code
            _schema.details = cls._schema_api_error_read.details
            _schema.innererror = cls._schema_api_error_read.innererror
            _schema.message = cls._schema_api_error_read.message
            _schema.target = cls._schema_api_error_read.target
            return

        cls._schema_api_error_read = _schema_api_error_read = AAZObjectType(
            flags={"read_only": True}
        )

        api_error_read = _schema_api_error_read
        api_error_read.code = AAZStrType()
        api_error_read.details = AAZListType()
        api_error_read.innererror = AAZObjectType()
        api_error_read.message = AAZStrType()
        api_error_read.target = AAZStrType()

        details = _schema_api_error_read.details
        details.Element = AAZObjectType()

        _element = _schema_api_error_read.details.Element
        _element.code = AAZStrType()
        _element.message = AAZStrType()
        _element.target = AAZStrType()

        innererror = _schema_api_error_read.innererror
        innererror.errordetail = AAZStrType()
        innererror.exceptiontype = AAZStrType()

        _schema.code = cls._schema_api_error_read.code
        _schema.details = cls._schema_api_error_read.details
        _schema.innererror = cls._schema_api_error_read.innererror
        _schema.message = cls._schema_api_error_read.message
        _schema.target = cls._schema_api_error_read.target

    _schema_key_for_disk_encryption_set_read = None

    @classmethod
    def _build_schema_key_for_disk_encryption_set_read(cls, _schema):
        if cls._schema_key_for_disk_encryption_set_read is not None:
            _schema.key_url = cls._schema_key_for_disk_encryption_set_read.key_url
            _schema.source_vault = cls._schema_key_for_disk_encryption_set_read.source_vault
            return

        cls._schema_key_for_disk_encryption_set_read = _schema_key_for_disk_encryption_set_read = AAZObjectType()

        key_for_disk_encryption_set_read = _schema_key_for_disk_encryption_set_read
        key_for_disk_encryption_set_read.key_url = AAZStrType(
            serialized_name="keyUrl",
            flags={"required": True},
        )
        key_for_disk_encryption_set_read.source_vault = AAZObjectType(
            serialized_name="sourceVault",
        )

        source_vault = _schema_key_for_disk_encryption_set_read.source_vault
        source_vault.id = AAZStrType()

        _schema.key_url = cls._schema_key_for_disk_encryption_set_read.key_url
        _schema.source_vault = cls._schema_key_for_disk_encryption_set_read.source_vault


__all__ = ["Wait"]



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
#
# Code generated by aaz-dev-tools
# --------------------------------------------------------------------------------------------

# pylint: skip-file
# flake8: noqa

from azure.cli.core.aaz import *


@register_command(
    "sql mi link failover",
    confirmation="This operation may cause data loss if the failover type is ForcedAllowDataLoss. Are you sure you want to proceed?",
)
class Failover(AAZCommand):
    """Performs requested failover type in this Managed Instance link.

    :example: Failover a Managed Instance link.
        az sql mi link failover -g testrg --mi testcl --name link1 --failover-type ForcedAllowDataLoss
    """

    _aaz_info = {
        "version": "2023-08-01-preview",
        "resources": [
            ["mgmt-plane", "/subscriptions/{}/resourcegroups/{}/providers/microsoft.sql/managedinstances/{}/distributedavailabilitygroups/{}/failover", "2023-08-01-preview"],
        ]
    }

    AZ_SUPPORT_NO_WAIT = True

    def _handler(self, command_args):
        super()._handler(command_args)
        return self.build_lro_poller(self._execute_operations, self._output)

    _args_schema = None

    @classmethod
    def _build_arguments_schema(cls, *args, **kwargs):
        if cls._args_schema is not None:
            return cls._args_schema
        cls._args_schema = super()._build_arguments_schema(*args, **kwargs)

        # define Arg Group ""

        _args_schema = cls._args_schema
        _args_schema.link_name = AAZStrArg(
            options=["-n", "--name", "--link-name"],
            help="Managed Instance link name.",
            required=True,
            id_part="child_name_1",
            fmt=AAZStrArgFormat(
                pattern="^[#a-zA-Z_][\\w@#$]*$",
                max_length=128,
                min_length=1,
            ),
        )
        _args_schema.managed_instance_name = AAZStrArg(
            options=["--mi", "--instance-name", "--managed-instance", "--managed-instance-name"],
            help="The name of the managed instance.",
            required=True,
            id_part="name",
            fmt=AAZStrArgFormat(
                pattern="^[a-z0-9]+(?:-[a-z0-9]+)*$",
                max_length=63,
                min_length=1,
            ),
        )
        _args_schema.resource_group = AAZResourceGroupNameArg(
            required=True,
        )

        # define Arg Group "Parameters"

        _args_schema = cls._args_schema
        _args_schema.failover_type = AAZStrArg(
            options=["--failover-type"],
            arg_group="Parameters",
            help="The failover type, can be ForcedAllowDataLoss or Planned.",
            required=True,
            enum={"ForcedAllowDataLoss": "ForcedAllowDataLoss", "Planned": "Planned"},
        )
        return cls._args_schema

    def _execute_operations(self):
        self.pre_operations()
        yield self.DistributedAvailabilityGroupsFailover(ctx=self.ctx)()
        self.post_operations()

    @register_callback
    def pre_operations(self):
        pass

    @register_callback
    def post_operations(self):
        pass

    def _output(self, *args, **kwargs):
        result = self.deserialize_output(self.ctx.vars.instance, client_flatten=True)
        return result

    class DistributedAvailabilityGroupsFailover(AAZHttpOperation):
        CLIENT_TYPE = "MgmtClient"

        def __call__(self, *args, **kwargs):
            request = self.make_request()
            session = self.client.send_request(request=request, stream=False, **kwargs)
            if session.http_response.status_code in [202]:
                return self.client.build_lro_polling(
                    self.ctx.args.no_wait,
                    session,
                    self.on_200,
                    self.on_error,
                    lro_options={"final-state-via": "location"},
                    path_format_arguments=self.url_parameters,
                )
            if session.http_response.status_code in [200]:
                return self.client.build_lro_polling(
                    self.ctx.args.no_wait,
                    session,
                    self.on_200,
                    self.on_error,
                    lro_options={"final-state-via": "location"},
                    path_format_arguments=self.url_parameters,
                )

            return self.on_error(session.http_response)

        @property
        def url(self):
            return self.client.format_url(
                "/subscriptions/{subscriptionId}/resourceGroups/{resourceGroupName}/providers/Microsoft.Sql/managedInstances/{managedInstanceName}/distributedAvailabilityGroups/{distributedAvailabilityGroupName}/failover",
                **self.url_parameters
            )

        @property
        def method(self):
            return "POST"

        @property
        def error_format(self):
            return "MgmtErrorFormat"

        @property
        def url_parameters(self):
            parameters = {
                **self.serialize_url_param(
                    "distributedAvailabilityGroupName", self.ctx.args.link_name,
                    required=True,
                ),
                **self.serialize_url_param(
                    "managedInstanceName", self.ctx.args.managed_instance_name,
                    required=True,
                ),
                **self.serialize_url_param(
                    "resourceGroupName", self.ctx.args.resource_group,
                    required=True,
                ),
                **self.serialize_url_param(
                    "subscriptionId", self.ctx.subscription_id,
                    required=True,
                ),
            }
            return parameters

        @property
        def query_parameters(self):
            parameters = {
                **self.serialize_query_param(
                    "api-version", "2023-08-01-preview",
                    required=True,
                ),
            }
            return parameters

        @property
        def header_parameters(self):
            parameters = {
                **self.serialize_header_param(
                    "Content-Type", "application/json",
                ),
                **self.serialize_header_param(
                    "Accept", "application/json",
                ),
            }
            return parameters

        @property
        def content(self):
            _content_value, _builder = self.new_content_builder(
                self.ctx.args,
                typ=AAZObjectType,
                typ_kwargs={"flags": {"required": True, "client_flatten": True}}
            )
            _builder.set_prop("failoverType", AAZStrType, ".failover_type", typ_kwargs={"flags": {"required": True}})

            return self.serialize_content(_content_value)

        def on_200(self, session):
            data = self.deserialize_http_content(session)
            self.ctx.set_var(
                "instance",
                data,
                schema_builder=self._build_schema_on_200
            )

        _schema_on_200 = None

        @classmethod
        def _build_schema_on_200(cls):
            if cls._schema_on_200 is not None:
                return cls._schema_on_200

            cls._schema_on_200 = AAZObjectType()

            _schema_on_200 = cls._schema_on_200
            _schema_on_200.id = AAZStrType(
                flags={"read_only": True},
            )
            _schema_on_200.name = AAZStrType(
                flags={"read_only": True},
            )
            _schema_on_200.properties = AAZObjectType(
                flags={"client_flatten": True},
            )
            _schema_on_200.type = AAZStrType(
                flags={"read_only": True},
            )

            properties = cls._schema_on_200.properties
            properties.databases = AAZListType()
            properties.distributed_availability_group_id = AAZStrType(
                serialized_name="distributedAvailabilityGroupId",
                flags={"read_only": True},
            )
            properties.distributed_availability_group_name = AAZStrType(
                serialized_name="distributedAvailabilityGroupName",
                flags={"read_only": True},
            )
            properties.failover_mode = AAZStrType(
                serialized_name="failoverMode",
            )
            properties.instance_availability_group_name = AAZStrType(
                serialized_name="instanceAvailabilityGroupName",
            )
            properties.instance_link_role = AAZStrType(
                serialized_name="instanceLinkRole",
            )
            properties.partner_availability_group_name = AAZStrType(
                serialized_name="partnerAvailabilityGroupName",
            )
            properties.partner_endpoint = AAZStrType(
                serialized_name="partnerEndpoint",
            )
            properties.partner_link_role = AAZStrType(
                serialized_name="partnerLinkRole",
                flags={"read_only": True},
            )
            properties.replication_mode = AAZStrType(
                serialized_name="replicationMode",
            )
            properties.seeding_mode = AAZStrType(
                serialized_name="seedingMode",
            )

            databases = cls._schema_on_200.properties.databases
            databases.Element = AAZObjectType()

            _element = cls._schema_on_200.properties.databases.Element
            _element.connected_state = AAZStrType(
                serialized_name="connectedState",
                flags={"read_only": True},
            )
            _element.database_name = AAZStrType(
                serialized_name="databaseName",
            )
            _element.instance_redo_replication_lag_seconds = AAZIntType(
                serialized_name="instanceRedoReplicationLagSeconds",
                flags={"read_only": True},
            )
            _element.instance_replica_id = AAZStrType(
                serialized_name="instanceReplicaId",
                flags={"read_only": True},
            )
            _element.instance_send_replication_lag_seconds = AAZIntType(
                serialized_name="instanceSendReplicationLagSeconds",
                flags={"read_only": True},
            )
            _element.last_backup_lsn = AAZStrType(
                serialized_name="lastBackupLsn",
                flags={"read_only": True},
            )
            _element.last_backup_time = AAZStrType(
                serialized_name="lastBackupTime",
                flags={"read_only": True},
            )
            _element.last_commit_lsn = AAZStrType(
                serialized_name="lastCommitLsn",
                flags={"read_only": True},
            )
            _element.last_commit_time = AAZStrType(
                serialized_name="lastCommitTime",
                flags={"read_only": True},
            )
            _element.last_hardened_lsn = AAZStrType(
                serialized_name="lastHardenedLsn",
                flags={"read_only": True},
            )
            _element.last_hardened_time = AAZStrType(
                serialized_name="lastHardenedTime",
                flags={"read_only": True},
            )
            _element.last_received_lsn = AAZStrType(
                serialized_name="lastReceivedLsn",
                flags={"read_only": True},
            )
            _element.last_received_time = AAZStrType(
                serialized_name="lastReceivedTime",
                flags={"read_only": True},
            )
            _element.last_sent_lsn = AAZStrType(
                serialized_name="lastSentLsn",
                flags={"read_only": True},
            )
            _element.last_sent_time = AAZStrType(
                serialized_name="lastSentTime",
                flags={"read_only": True},
            )
            _element.most_recent_link_error = AAZStrType(
                serialized_name="mostRecentLinkError",
                flags={"read_only": True},
            )
            _element.partner_auth_cert_validity = AAZObjectType(
                serialized_name="partnerAuthCertValidity",
                flags={"read_only": True},
            )
            _element.partner_replica_id = AAZStrType(
                serialized_name="partnerReplicaId",
                flags={"read_only": True},
            )
            _element.replica_state = AAZStrType(
                serialized_name="replicaState",
                flags={"read_only": True},
            )
            _element.seeding_progress = AAZStrType(
                serialized_name="seedingProgress",
                flags={"read_only": True},
            )
            _element.synchronization_health = AAZStrType(
                serialized_name="synchronizationHealth",
                flags={"read_only": True},
            )

            partner_auth_cert_validity = cls._schema_on_200.properties.databases.Element.partner_auth_cert_validity
            partner_auth_cert_validity.certificate_name = AAZStrType(
                serialized_name="certificateName",
                flags={"read_only": True},
            )
            partner_auth_cert_validity.expiry_date = AAZStrType(
                serialized_name="expiryDate",
                flags={"read_only": True},
            )

            return cls._schema_on_200


class _FailoverHelper:
    """Helper class for Failover"""


__all__ = ["Failover"]



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
#
# Code generated by aaz-dev-tools
# --------------------------------------------------------------------------------------------

# pylint: skip-file
# flake8: noqa

from azure.cli.core.aaz import *


class Update(AAZCommand):
    """Update a gallery image version.
    """

    _aaz_info = {
        "version": "2024-03-03",
        "resources": [
            ["mgmt-plane", "/subscriptions/{}/resourcegroups/{}/providers/microsoft.compute/galleries/{}/images/{}/versions/{}", "2024-03-03"],
        ]
    }

    AZ_SUPPORT_NO_WAIT = True

    AZ_SUPPORT_GENERIC_UPDATE = True

    def _handler(self, command_args):
        super()._handler(command_args)
        return self.build_lro_poller(self._execute_operations, self._output)

    _args_schema = None

    @classmethod
    def _build_arguments_schema(cls, *args, **kwargs):
        if cls._args_schema is not None:
            return cls._args_schema
        cls._args_schema = super()._build_arguments_schema(*args, **kwargs)

        # define Arg Group ""

        _args_schema = cls._args_schema
        _args_schema.gallery_image_definition = AAZStrArg(
            options=["-i", "--gallery-image-name", "--gallery-image-definition"],
            help="The name of the gallery image definition in which the Image Version resides.",
            required=True,
            id_part="child_name_1",
        )
        _args_schema.gallery_image_version_name = AAZStrArg(
            options=["-e", "--gallery-image-version", "--gallery-image-version-name"],
            help="The name of the gallery image version to be deleted.",
            required=True,
            id_part="child_name_2",
        )
        _args_schema.gallery_name = AAZStrArg(
            options=["-r", "--gallery-name"],
            help="The name of the Shared Image Gallery in which the Image Definition resides.",
            required=True,
            id_part="name",
        )
        _args_schema.resource_group = AAZResourceGroupNameArg(
            required=True,
        )

        # define Arg Group "GalleryImageVersion"

        _args_schema = cls._args_schema
        _args_schema.location = AAZResourceLocationArg(
            arg_group="GalleryImageVersion",
            help="Resource location",
            fmt=AAZResourceLocationArgFormat(
                resource_group_arg="resource_group",
            ),
        )
        _args_schema.tags = AAZDictArg(
            options=["--tags"],
            arg_group="GalleryImageVersion",
            help="Resource tags",
            nullable=True,
        )

        tags = cls._args_schema.tags
        tags.Element = AAZStrArg(
            nullable=True,
        )

        # define Arg Group "Properties"

        _args_schema = cls._args_schema
        _args_schema.publishing_profile = AAZObjectArg(
            options=["--publishing-profile"],
            arg_group="Properties",
            help="The publishing profile of a gallery image Version.",
            nullable=True,
        )
        _args_schema.restore = AAZBoolArg(
            options=["--restore"],
            arg_group="Properties",
            help="Indicates if this is a soft-delete resource restoration request.",
            nullable=True,
        )
        _args_schema.safety_profile = AAZObjectArg(
            options=["--safety-profile"],
            arg_group="Properties",
            help="This is the safety profile of the Gallery Image Version.",
            nullable=True,
        )
        _args_schema.security_profile = AAZObjectArg(
            options=["--security-profile"],
            arg_group="Properties",
            help="The security profile of a gallery image version",
            nullable=True,
        )
        _args_schema.storage_profile = AAZObjectArg(
            options=["--storage-profile"],
            arg_group="Properties",
            help="This is the storage profile of a Gallery Image Version.",
        )

        publishing_profile = cls._args_schema.publishing_profile
        publishing_profile.end_of_life_date = AAZDateTimeArg(
            options=["end-of-life-date"],
            help="The end of life date of the gallery image version. This property can be used for decommissioning purposes. This property is updatable.",
            nullable=True,
        )
        publishing_profile.exclude_from_latest = AAZBoolArg(
            options=["exclude-from-latest"],
            help="If set to true, Virtual Machines deployed from the latest version of the Image Definition won't use this Image Version.",
            nullable=True,
        )
        publishing_profile.replica_count = AAZIntArg(
            options=["replica-count"],
            help="The number of replicas of the Image Version to be created per region. This property would take effect for a region when regionalReplicaCount is not specified. This property is updatable.",
            nullable=True,
        )
        publishing_profile.replication_mode = AAZStrArg(
            options=["replication-mode"],
            help="Optional parameter which specifies the mode to be used for replication. This property is not updatable.",
            nullable=True,
            enum={"Full": "Full", "Shallow": "Shallow"},
        )
        publishing_profile.storage_account_type = AAZStrArg(
            options=["storage-account-type"],
            help="Specifies the storage account type to be used to store the image. This property is not updatable.",
            nullable=True,
            enum={"PremiumV2_LRS": "PremiumV2_LRS", "Premium_LRS": "Premium_LRS", "Standard_LRS": "Standard_LRS", "Standard_ZRS": "Standard_ZRS"},
        )
        publishing_profile.target_extended_locations = AAZListArg(
            options=["target-extended-locations"],
            help="The target extended locations where the Image Version is going to be replicated to. This property is updatable.",
            nullable=True,
        )
        publishing_profile.target_regions = AAZListArg(
            options=["target-regions"],
            help="The target regions where the Image Version is going to be replicated to. This property is updatable.",
            nullable=True,
        )

        target_extended_locations = cls._args_schema.publishing_profile.target_extended_locations
        target_extended_locations.Element = AAZObjectArg(
            nullable=True,
        )

        _element = cls._args_schema.publishing_profile.target_extended_locations.Element
        _element.encryption = AAZObjectArg(
            options=["encryption"],
            nullable=True,
        )
        cls._build_args_encryption_images_update(_element.encryption)
        _element.extended_location = AAZObjectArg(
            options=["extended-location"],
            help="The name of the extended location.",
            nullable=True,
        )
        _element.extended_location_replica_count = AAZIntArg(
            options=["extended-location-replica-count"],
            help="The number of replicas of the Image Version to be created per extended location. This property is updatable.",
            nullable=True,
        )
        _element.name = AAZStrArg(
            options=["name"],
            help="The name of the region.",
            nullable=True,
        )
        _element.storage_account_type = AAZStrArg(
            options=["storage-account-type"],
            help="Specifies the storage account type to be used to store the image. This property is not updatable.",
            nullable=True,
            enum={"Premium_LRS": "Premium_LRS", "StandardSSD_LRS": "StandardSSD_LRS", "Standard_LRS": "Standard_LRS", "Standard_ZRS": "Standard_ZRS"},
        )

        extended_location = cls._args_schema.publishing_profile.target_extended_locations.Element.extended_location
        extended_location.name = AAZStrArg(
            options=["name"],
            nullable=True,
        )
        extended_location.type = AAZStrArg(
            options=["type"],
            help="It is type of the extended location.",
            nullable=True,
            enum={"EdgeZone": "EdgeZone", "Unknown": "Unknown"},
        )

        target_regions = cls._args_schema.publishing_profile.target_regions
        target_regions.Element = AAZObjectArg(
            nullable=True,
        )

        _element = cls._args_schema.publishing_profile.target_regions.Element
        _element.additional_replica_sets = AAZListArg(
            options=["additional-replica-sets"],
            help="List of storage sku with replica count to create direct drive replicas.",
            nullable=True,
        )
        _element.encryption = AAZObjectArg(
            options=["encryption"],
            help="Optional. Allows users to provide customer managed keys for encrypting the OS and data disks in the gallery artifact.",
            nullable=True,
        )
        cls._build_args_encryption_images_update(_element.encryption)
        _element.exclude_from_latest = AAZBoolArg(
            options=["exclude-from-latest"],
            help="Contains the flag setting to hide an image when users specify version='latest'",
            nullable=True,
        )
        _element.name = AAZStrArg(
            options=["name"],
            help="The name of the region.",
        )
        _element.regional_replica_count = AAZIntArg(
            options=["regional-replica-count"],
            help="The number of replicas of the Image Version to be created per region. This property is updatable.",
            nullable=True,
        )
        _element.storage_account_type = AAZStrArg(
            options=["storage-account-type"],
            help="Specifies the storage account type to be used to store the image. This property is not updatable.",
            nullable=True,
            enum={"PremiumV2_LRS": "PremiumV2_LRS", "Premium_LRS": "Premium_LRS", "Standard_LRS": "Standard_LRS", "Standard_ZRS": "Standard_ZRS"},
        )

        additional_replica_sets = cls._args_schema.publishing_profile.target_regions.Element.additional_replica_sets
        additional_replica_sets.Element = AAZObjectArg(
            nullable=True,
        )

        _element = cls._args_schema.publishing_profile.target_regions.Element.additional_replica_sets.Element
        _element.regional_replica_count = AAZIntArg(
            options=["regional-replica-count"],
            help="The number of direct drive replicas of the Image Version to be created.This Property is updatable",
            nullable=True,
        )
        _element.storage_account_type = AAZStrArg(
            options=["storage-account-type"],
            help="Specifies the storage account type to be used to create the direct drive replicas",
            nullable=True,
            enum={"PremiumV2_LRS": "PremiumV2_LRS", "Premium_LRS": "Premium_LRS", "Standard_LRS": "Standard_LRS", "Standard_ZRS": "Standard_ZRS"},
        )

        safety_profile = cls._args_schema.safety_profile
        safety_profile.allow_deletion_of_replicated_locations = AAZBoolArg(
            options=["allow-deletion-of-replicated-locations"],
            help="Indicates whether or not removing this Gallery Image Version from replicated regions is allowed.",
            nullable=True,
        )
        safety_profile.block_deletion_before_end_of_life = AAZBoolArg(
            options=["block-deletion-before-end-of-life"],
            help="Indicates whether or not the deletion is blocked for this Gallery Image Version if its End Of Life has not expired.",
            nullable=True,
        )

        security_profile = cls._args_schema.security_profile
        security_profile.uefi_settings = AAZObjectArg(
            options=["uefi-settings"],
            help="Contains UEFI settings for the image version.",
            nullable=True,
        )

        uefi_settings = cls._args_schema.security_profile.uefi_settings
        uefi_settings.additional_signatures = AAZObjectArg(
            options=["additional-signatures"],
            help="Additional UEFI key signatures that will be added to the image in addition to the signature templates",
            nullable=True,
        )
        uefi_settings.signature_template_names = AAZListArg(
            options=["signature-template-names"],
            help="The name of the template(s) that contains default UEFI key signatures that will be added to the image.",
            nullable=True,
        )

        additional_signatures = cls._args_schema.security_profile.uefi_settings.additional_signatures
        additional_signatures.db = AAZListArg(
            options=["db"],
            help="The database of UEFI keys for this image version.",
            nullable=True,
        )
        additional_signatures.dbx = AAZListArg(
            options=["dbx"],
            help="The database of revoked UEFI keys for this image version.",
            nullable=True,
        )
        additional_signatures.kek = AAZListArg(
            options=["kek"],
            help="The Key Encryption Keys of this image version.",
            nullable=True,
        )
        additional_signatures.pk = AAZObjectArg(
            options=["pk"],
            help="The Platform Key of this image version.",
            nullable=True,
        )
        cls._build_args_uefi_key_update(additional_signatures.pk)

        db = cls._args_schema.security_profile.uefi_settings.additional_signatures.db
        db.Element = AAZObjectArg(
            nullable=True,
        )
        cls._build_args_uefi_key_update(db.Element)

        dbx = cls._args_schema.security_profile.uefi_settings.additional_signatures.dbx
        dbx.Element = AAZObjectArg(
            nullable=True,
        )
        cls._build_args_uefi_key_update(dbx.Element)

        kek = cls._args_schema.security_profile.uefi_settings.additional_signatures.kek
        kek.Element = AAZObjectArg(
            nullable=True,
        )
        cls._build_args_uefi_key_update(kek.Element)

        signature_template_names = cls._args_schema.security_profile.uefi_settings.signature_template_names
        signature_template_names.Element = AAZStrArg(
            nullable=True,
            enum={"MicrosoftUefiCertificateAuthorityTemplate": "MicrosoftUefiCertificateAuthorityTemplate", "MicrosoftWindowsTemplate": "MicrosoftWindowsTemplate", "NoSignatureTemplate": "NoSignatureTemplate"},
        )

        storage_profile = cls._args_schema.storage_profile
        storage_profile.data_disk_images = AAZListArg(
            options=["data-disk-images"],
            help="A list of data disk images.",
            nullable=True,
        )
        storage_profile.os_disk_image = AAZObjectArg(
            options=["os-disk-image"],
            help="This is the OS disk image.",
            nullable=True,
        )
        storage_profile.source = AAZObjectArg(
            options=["source"],
            help="The source of the gallery artifact version.",
            nullable=True,
        )

        data_disk_images = cls._args_schema.storage_profile.data_disk_images
        data_disk_images.Element = AAZObjectArg(
            nullable=True,
        )

        _element = cls._args_schema.storage_profile.data_disk_images.Element
        _element.host_caching = AAZStrArg(
            options=["host-caching"],
            help="The host caching of the disk. Valid values are 'None', 'ReadOnly', and 'ReadWrite'",
            nullable=True,
            enum={"None": "None", "ReadOnly": "ReadOnly", "ReadWrite": "ReadWrite"},
        )
        _element.lun = AAZIntArg(
            options=["lun"],
            help="This property specifies the logical unit number of the data disk. This value is used to identify data disks within the Virtual Machine and therefore must be unique for each data disk attached to the Virtual Machine.",
        )
        _element.source = AAZObjectArg(
            options=["source"],
            nullable=True,
        )
        cls._build_args_gallery_disk_image_source_update(_element.source)

        os_disk_image = cls._args_schema.storage_profile.os_disk_image
        os_disk_image.host_caching = AAZStrArg(
            options=["host-caching"],
            help="The host caching of the disk. Valid values are 'None', 'ReadOnly', and 'ReadWrite'",
            nullable=True,
            enum={"None": "None", "ReadOnly": "ReadOnly", "ReadWrite": "ReadWrite"},
        )
        os_disk_image.source = AAZObjectArg(
            options=["source"],
            help="The source for the disk image.",
            nullable=True,
        )
        cls._build_args_gallery_disk_image_source_update(os_disk_image.source)

        source = cls._args_schema.storage_profile.source
        source.community_gallery_image_id = AAZStrArg(
            options=["community-gallery-image-id"],
            help="The resource Id of the source Community Gallery Image.  Only required when using Community Gallery Image as a source.",
            nullable=True,
        )
        source.id = AAZResourceIdArg(
            options=["id"],
            help="The id of the gallery artifact version source. Can specify a disk uri, snapshot uri, user image or storage account resource.",
            nullable=True,
        )
        source.virtual_machine_id = AAZResourceIdArg(
            options=["virtual-machine-id"],
            help="The resource Id of the source virtual machine.  Only required when capturing a virtual machine to source this Gallery Image Version.",
            nullable=True,
        )
        return cls._args_schema

    _args_encryption_images_update = None

    @classmethod
    def _build_args_encryption_images_update(cls, _schema):
        if cls._args_encryption_images_update is not None:
            _schema.data_disk_images = cls._args_encryption_images_update.data_disk_images
            _schema.os_disk_image = cls._args_encryption_images_update.os_disk_image
            return

        cls._args_encryption_images_update = AAZObjectArg(
            nullable=True,
        )

        encryption_images_update = cls._args_encryption_images_update
        encryption_images_update.data_disk_images = AAZListArg(
            options=["data-disk-images"],
            help="A list of encryption specifications for data disk images.",
            nullable=True,
        )
        encryption_images_update.os_disk_image = AAZObjectArg(
            options=["os-disk-image"],
            help="Contains encryption settings for an OS disk image.",
            nullable=True,
        )

        data_disk_images = cls._args_encryption_images_update.data_disk_images
        data_disk_images.Element = AAZObjectArg(
            nullable=True,
        )

        _element = cls._args_encryption_images_update.data_disk_images.Element
        _element.disk_encryption_set_id = AAZStrArg(
            options=["disk-encryption-set-id"],
            help="A relative URI containing the resource ID of the disk encryption set.",
            nullable=True,
        )
        _element.lun = AAZIntArg(
            options=["lun"],
            help="This property specifies the logical unit number of the data disk. This value is used to identify data disks within the Virtual Machine and therefore must be unique for each data disk attached to the Virtual Machine.",
        )

        os_disk_image = cls._args_encryption_images_update.os_disk_image
        os_disk_image.disk_encryption_set_id = AAZStrArg(
            options=["disk-encryption-set-id"],
            help="A relative URI containing the resource ID of the disk encryption set.",
            nullable=True,
        )
        os_disk_image.security_profile = AAZObjectArg(
            options=["security-profile"],
            help="This property specifies the security profile of an OS disk image.",
            nullable=True,
        )

        security_profile = cls._args_encryption_images_update.os_disk_image.security_profile
        security_profile.confidential_vm_encryption_type = AAZStrArg(
            options=["confidential-vm-encryption-type"],
            help="confidential VM encryption types",
            nullable=True,
            enum={"EncryptedVMGuestStateOnlyWithPmk": "EncryptedVMGuestStateOnlyWithPmk", "EncryptedWithCmk": "EncryptedWithCmk", "EncryptedWithPmk": "EncryptedWithPmk", "NonPersistedTPM": "NonPersistedTPM"},
        )
        security_profile.secure_vm_disk_encryption_set_id = AAZStrArg(
            options=["secure-vm-disk-encryption-set-id"],
            help="secure VM disk encryption set id",
            nullable=True,
        )

        _schema.data_disk_images = cls._args_encryption_images_update.data_disk_images
        _schema.os_disk_image = cls._args_encryption_images_update.os_disk_image

    _args_gallery_disk_image_source_update = None

    @classmethod
    def _build_args_gallery_disk_image_source_update(cls, _schema):
        if cls._args_gallery_disk_image_source_update is not None:
            _schema.id = cls._args_gallery_disk_image_source_update.id
            _schema.storage_account_id = cls._args_gallery_disk_image_source_update.storage_account_id
            _schema.uri = cls._args_gallery_disk_image_source_update.uri
            return

        cls._args_gallery_disk_image_source_update = AAZObjectArg(
            nullable=True,
        )

        gallery_disk_image_source_update = cls._args_gallery_disk_image_source_update
        gallery_disk_image_source_update.id = AAZResourceIdArg(
            options=["id"],
            help="The id of the gallery artifact version source. Can specify a disk uri, snapshot uri, user image or storage account resource.",
            nullable=True,
        )
        gallery_disk_image_source_update.storage_account_id = AAZResourceIdArg(
            options=["storage-account-id"],
            help="The Storage Account Id that contains the vhd blob being used as a source for this artifact version.",
            nullable=True,
        )
        gallery_disk_image_source_update.uri = AAZStrArg(
            options=["uri"],
            help="The uri of the gallery artifact version source. Currently used to specify vhd/blob source.",
            nullable=True,
        )

        _schema.id = cls._args_gallery_disk_image_source_update.id
        _schema.storage_account_id = cls._args_gallery_disk_image_source_update.storage_account_id
        _schema.uri = cls._args_gallery_disk_image_source_update.uri

    _args_uefi_key_update = None

    @classmethod
    def _build_args_uefi_key_update(cls, _schema):
        if cls._args_uefi_key_update is not None:
            _schema.type = cls._args_uefi_key_update.type
            _schema.value = cls._args_uefi_key_update.value
            return

        cls._args_uefi_key_update = AAZObjectArg(
            nullable=True,
        )

        uefi_key_update = cls._args_uefi_key_update
        uefi_key_update.type = AAZStrArg(
            options=["type"],
            help="The type of key signature.",
            nullable=True,
            enum={"sha256": "sha256", "x509": "x509"},
        )
        uefi_key_update.value = AAZListArg(
            options=["value"],
            help="The value of the key signature.",
            nullable=True,
        )

        value = cls._args_uefi_key_update.value
        value.Element = AAZStrArg(
            nullable=True,
        )

        _schema.type = cls._args_uefi_key_update.type
        _schema.value = cls._args_uefi_key_update.value

    def _execute_operations(self):
        self.pre_operations()
        self.GalleryImageVersionsGet(ctx=self.ctx)()
        self.pre_instance_update(self.ctx.vars.instance)
        self.InstanceUpdateByJson(ctx=self.ctx)()
        self.InstanceUpdateByGeneric(ctx=self.ctx)()
        self.post_instance_update(self.ctx.vars.instance)
        yield self.GalleryImageVersionsCreateOrUpdate(ctx=self.ctx)()
        self.post_operations()

    @register_callback
    def pre_operations(self):
        pass

    @register_callback
    def post_operations(self):
        pass

    @register_callback
    def pre_instance_update(self, instance):
        pass

    @register_callback
    def post_instance_update(self, instance):
        pass

    def _output(self, *args, **kwargs):
        result = self.deserialize_output(self.ctx.vars.instance, client_flatten=True)
        return result

    class GalleryImageVersionsGet(AAZHttpOperation):
        CLIENT_TYPE = "MgmtClient"

        def __call__(self, *args, **kwargs):
            request = self.make_request()
            session = self.client.send_request(request=request, stream=False, **kwargs)
            if session.http_response.status_code in [200]:
                return self.on_200(session)

            return self.on_error(session.http_response)

        @property
        def url(self):
            return self.client.format_url(
                "/subscriptions/{subscriptionId}/resourceGroups/{resourceGroupName}/providers/Microsoft.Compute/galleries/{galleryName}/images/{galleryImageName}/versions/{galleryImageVersionName}",
                **self.url_parameters
            )

        @property
        def method(self):
            return "GET"

        @property
        def error_format(self):
            return "ODataV4Format"

        @property
        def url_parameters(self):
            parameters = {
                **self.serialize_url_param(
                    "galleryImageName", self.ctx.args.gallery_image_definition,
                    required=True,
                ),
                **self.serialize_url_param(
                    "galleryImageVersionName", self.ctx.args.gallery_image_version_name,
                    required=True,
                ),
                **self.serialize_url_param(
                    "galleryName", self.ctx.args.gallery_name,
                    required=True,
                ),
                **self.serialize_url_param(
                    "resourceGroupName", self.ctx.args.resource_group,
                    required=True,
                ),
                **self.serialize_url_param(
                    "subscriptionId", self.ctx.subscription_id,
                    required=True,
                ),
            }
            return parameters

        @property
        def query_parameters(self):
            parameters = {
                **self.serialize_query_param(
                    "api-version", "2024-03-03",
                    required=True,
                ),
            }
            return parameters

        @property
        def header_parameters(self):
            parameters = {
                **self.serialize_header_param(
                    "Accept", "application/json",
                ),
            }
            return parameters

        def on_200(self, session):
            data = self.deserialize_http_content(session)
            self.ctx.set_var(
                "instance",
                data,
                schema_builder=self._build_schema_on_200
            )

        _schema_on_200 = None

        @classmethod
        def _build_schema_on_200(cls):
            if cls._schema_on_200 is not None:
                return cls._schema_on_200

            cls._schema_on_200 = AAZObjectType()
            _UpdateHelper._build_schema_gallery_image_version_read(cls._schema_on_200)

            return cls._schema_on_200

    class GalleryImageVersionsCreateOrUpdate(AAZHttpOperation):
        CLIENT_TYPE = "MgmtClient"

        def __call__(self, *args, **kwargs):
            request = self.make_request()
            session = self.client.send_request(request=request, stream=False, **kwargs)
            if session.http_response.status_code in [202]:
                return self.client.build_lro_polling(
                    self.ctx.args.no_wait,
                    session,
                    self.on_200_201,
                    self.on_error,
                    lro_options={"final-state-via": "azure-async-operation"},
                    path_format_arguments=self.url_parameters,
                )
            if session.http_response.status_code in [200, 201]:
                return self.client.build_lro_polling(
                    self.ctx.args.no_wait,
                    session,
                    self.on_200_201,
                    self.on_error,
                    lro_options={"final-state-via": "azure-async-operation"},
                    path_format_arguments=self.url_parameters,
                )

            return self.on_error(session.http_response)

        @property
        def url(self):
            return self.client.format_url(
                "/subscriptions/{subscriptionId}/resourceGroups/{resourceGroupName}/providers/Microsoft.Compute/galleries/{galleryName}/images/{galleryImageName}/versions/{galleryImageVersionName}",
                **self.url_parameters
            )

        @property
        def method(self):
            return "PUT"

        @property
        def error_format(self):
            return "ODataV4Format"

        @property
        def url_parameters(self):
            parameters = {
                **self.serialize_url_param(
                    "galleryImageName", self.ctx.args.gallery_image_definition,
                    required=True,
                ),
                **self.serialize_url_param(
                    "galleryImageVersionName", self.ctx.args.gallery_image_version_name,
                    required=True,
                ),
                **self.serialize_url_param(
                    "galleryName", self.ctx.args.gallery_name,
                    required=True,
                ),
                **self.serialize_url_param(
                    "resourceGroupName", self.ctx.args.resource_group,
                    required=True,
                ),
                **self.serialize_url_param(
                    "subscriptionId", self.ctx.subscription_id,
                    required=True,
                ),
            }
            return parameters

        @property
        def query_parameters(self):
            parameters = {
                **self.serialize_query_param(
                    "api-version", "2024-03-03",
                    required=True,
                ),
            }
            return parameters

        @property
        def header_parameters(self):
            parameters = {
                **self.serialize_header_param(
                    "Content-Type", "application/json",
                ),
                **self.serialize_header_param(
                    "Accept", "application/json",
                ),
            }
            return parameters

        @property
        def content(self):
            _content_value, _builder = self.new_content_builder(
                self.ctx.args,
                value=self.ctx.vars.instance,
            )

            return self.serialize_content(_content_value)

        def on_200_201(self, session):
            data = self.deserialize_http_content(session)
            self.ctx.set_var(
                "instance",
                data,
                schema_builder=self._build_schema_on_200_201
            )

        _schema_on_200_201 = None

        @classmethod
        def _build_schema_on_200_201(cls):
            if cls._schema_on_200_201 is not None:
                return cls._schema_on_200_201

            cls._schema_on_200_201 = AAZObjectType()
            _UpdateHelper._build_schema_gallery_image_version_read(cls._schema_on_200_201)

            return cls._schema_on_200_201

    class InstanceUpdateByJson(AAZJsonInstanceUpdateOperation):

        def __call__(self, *args, **kwargs):
            self._update_instance(self.ctx.vars.instance)

        def _update_instance(self, instance):
            _instance_value, _builder = self.new_content_builder(
                self.ctx.args,
                value=instance,
                typ=AAZObjectType
            )
            _builder.set_prop("location", AAZStrType, ".location", typ_kwargs={"flags": {"required": True}})
            _builder.set_prop("properties", AAZObjectType, typ_kwargs={"flags": {"client_flatten": True}})
            _builder.set_prop("tags", AAZDictType, ".tags")

            properties = _builder.get(".properties")
            if properties is not None:
                properties.set_prop("publishingProfile", AAZObjectType, ".publishing_profile")
                properties.set_prop("restore", AAZBoolType, ".restore")
                properties.set_prop("safetyProfile", AAZObjectType, ".safety_profile")
                properties.set_prop("securityProfile", AAZObjectType, ".security_profile")
                properties.set_prop("storageProfile", AAZObjectType, ".storage_profile", typ_kwargs={"flags": {"required": True}})

            publishing_profile = _builder.get(".properties.publishingProfile")
            if publishing_profile is not None:
                publishing_profile.set_prop("endOfLifeDate", AAZStrType, ".end_of_life_date")
                publishing_profile.set_prop("excludeFromLatest", AAZBoolType, ".exclude_from_latest")
                publishing_profile.set_prop("replicaCount", AAZIntType, ".replica_count")
                publishing_profile.set_prop("replicationMode", AAZStrType, ".replication_mode")
                publishing_profile.set_prop("storageAccountType", AAZStrType, ".storage_account_type")
                publishing_profile.set_prop("targetExtendedLocations", AAZListType, ".target_extended_locations")
                publishing_profile.set_prop("targetRegions", AAZListType, ".target_regions")

            target_extended_locations = _builder.get(".properties.publishingProfile.targetExtendedLocations")
            if target_extended_locations is not None:
                target_extended_locations.set_elements(AAZObjectType, ".")

            _elements = _builder.get(".properties.publishingProfile.targetExtendedLocations[]")
            if _elements is not None:
                _UpdateHelper._build_schema_encryption_images_update(_elements.set_prop("encryption", AAZObjectType, ".encryption"))
                _elements.set_prop("extendedLocation", AAZObjectType, ".extended_location")
                _elements.set_prop("extendedLocationReplicaCount", AAZIntType, ".extended_location_replica_count")
                _elements.set_prop("name", AAZStrType, ".name")
                _elements.set_prop("storageAccountType", AAZStrType, ".storage_account_type")

            extended_location = _builder.get(".properties.publishingProfile.targetExtendedLocations[].extendedLocation")
            if extended_location is not None:
                extended_location.set_prop("name", AAZStrType, ".name")
                extended_location.set_prop("type", AAZStrType, ".type")

            target_regions = _builder.get(".properties.publishingProfile.targetRegions")
            if target_regions is not None:
                target_regions.set_elements(AAZObjectType, ".")

            _elements = _builder.get(".properties.publishingProfile.targetRegions[]")
            if _elements is not None:
                _elements.set_prop("additionalReplicaSets", AAZListType, ".additional_replica_sets")
                _UpdateHelper._build_schema_encryption_images_update(_elements.set_prop("encryption", AAZObjectType, ".encryption"))
                _elements.set_prop("excludeFromLatest", AAZBoolType, ".exclude_from_latest")
                _elements.set_prop("name", AAZStrType, ".name", typ_kwargs={"flags": {"required": True}})
                _elements.set_prop("regionalReplicaCount", AAZIntType, ".regional_replica_count")
                _elements.set_prop("storageAccountType", AAZStrType, ".storage_account_type")

            additional_replica_sets = _builder.get(".properties.publishingProfile.targetRegions[].additionalReplicaSets")
            if additional_replica_sets is not None:
                additional_replica_sets.set_elements(AAZObjectType, ".")

            _elements = _builder.get(".properties.publishingProfile.targetRegions[].additionalReplicaSets[]")
            if _elements is not None:
                _elements.set_prop("regionalReplicaCount", AAZIntType, ".regional_replica_count")
                _elements.set_prop("storageAccountType", AAZStrType, ".storage_account_type")

            safety_profile = _builder.get(".properties.safetyProfile")
            if safety_profile is not None:
                safety_profile.set_prop("allowDeletionOfReplicatedLocations", AAZBoolType, ".allow_deletion_of_replicated_locations")
                safety_profile.set_prop("blockDeletionBeforeEndOfLife", AAZBoolType, ".block_deletion_before_end_of_life")

            security_profile = _builder.get(".properties.securityProfile")
            if security_profile is not None:
                security_profile.set_prop("uefiSettings", AAZObjectType, ".uefi_settings")

            uefi_settings = _builder.get(".properties.securityProfile.uefiSettings")
            if uefi_settings is not None:
                uefi_settings.set_prop("additionalSignatures", AAZObjectType, ".additional_signatures")
                uefi_settings.set_prop("signatureTemplateNames", AAZListType, ".signature_template_names")

            additional_signatures = _builder.get(".properties.securityProfile.uefiSettings.additionalSignatures")
            if additional_signatures is not None:
                additional_signatures.set_prop("db", AAZListType, ".db")
                additional_signatures.set_prop("dbx", AAZListType, ".dbx")
                additional_signatures.set_prop("kek", AAZListType, ".kek")
                _UpdateHelper._build_schema_uefi_key_update(additional_signatures.set_prop("pk", AAZObjectType, ".pk"))

            db = _builder.get(".properties.securityProfile.uefiSettings.additionalSignatures.db")
            if db is not None:
                _UpdateHelper._build_schema_uefi_key_update(db.set_elements(AAZObjectType, "."))

            dbx = _builder.get(".properties.securityProfile.uefiSettings.additionalSignatures.dbx")
            if dbx is not None:
                _UpdateHelper._build_schema_uefi_key_update(dbx.set_elements(AAZObjectType, "."))

            kek = _builder.get(".properties.securityProfile.uefiSettings.additionalSignatures.kek")
            if kek is not None:
                _UpdateHelper._build_schema_uefi_key_update(kek.set_elements(AAZObjectType, "."))

            signature_template_names = _builder.get(".properties.securityProfile.uefiSettings.signatureTemplateNames")
            if signature_template_names is not None:
                signature_template_names.set_elements(AAZStrType, ".")

            storage_profile = _builder.get(".properties.storageProfile")
            if storage_profile is not None:
                storage_profile.set_prop("dataDiskImages", AAZListType, ".data_disk_images")
                storage_profile.set_prop("osDiskImage", AAZObjectType, ".os_disk_image")
                storage_profile.set_prop("source", AAZObjectType, ".source")

            data_disk_images = _builder.get(".properties.storageProfile.dataDiskImages")
            if data_disk_images is not None:
                data_disk_images.set_elements(AAZObjectType, ".")

            _elements = _builder.get(".properties.storageProfile.dataDiskImages[]")
            if _elements is not None:
                _elements.set_prop("hostCaching", AAZStrType, ".host_caching")
                _elements.set_prop("lun", AAZIntType, ".lun", typ_kwargs={"flags": {"required": True}})
                _UpdateHelper._build_schema_gallery_disk_image_source_update(_elements.set_prop("source", AAZObjectType, ".source"))

            os_disk_image = _builder.get(".properties.storageProfile.osDiskImage")
            if os_disk_image is not None:
                os_disk_image.set_prop("hostCaching", AAZStrType, ".host_caching")
                _UpdateHelper._build_schema_gallery_disk_image_source_update(os_disk_image.set_prop("source", AAZObjectType, ".source"))

            source = _builder.get(".properties.storageProfile.source")
            if source is not None:
                source.set_prop("communityGalleryImageId", AAZStrType, ".community_gallery_image_id")
                source.set_prop("id", AAZStrType, ".id")
                source.set_prop("virtualMachineId", AAZStrType, ".virtual_machine_id")

            tags = _builder.get(".tags")
            if tags is not None:
                tags.set_elements(AAZStrType, ".")

            return _instance_value

    class InstanceUpdateByGeneric(AAZGenericInstanceUpdateOperation):

        def __call__(self, *args, **kwargs):
            self._update_instance_by_generic(
                self.ctx.vars.instance,
                self.ctx.generic_update_args
            )


class _UpdateHelper:
    """Helper class for Update"""

    @classmethod
    def _build_schema_encryption_images_update(cls, _builder):
        if _builder is None:
            return
        _builder.set_prop("dataDiskImages", AAZListType, ".data_disk_images")
        _builder.set_prop("osDiskImage", AAZObjectType, ".os_disk_image")

        data_disk_images = _builder.get(".dataDiskImages")
        if data_disk_images is not None:
            data_disk_images.set_elements(AAZObjectType, ".")

        _elements = _builder.get(".dataDiskImages[]")
        if _elements is not None:
            _elements.set_prop("diskEncryptionSetId", AAZStrType, ".disk_encryption_set_id")
            _elements.set_prop("lun", AAZIntType, ".lun", typ_kwargs={"flags": {"required": True}})

        os_disk_image = _builder.get(".osDiskImage")
        if os_disk_image is not None:
            os_disk_image.set_prop("diskEncryptionSetId", AAZStrType, ".disk_encryption_set_id")
            os_disk_image.set_prop("securityProfile", AAZObjectType, ".security_profile")

        security_profile = _builder.get(".osDiskImage.securityProfile")
        if security_profile is not None:
            security_profile.set_prop("confidentialVMEncryptionType", AAZStrType, ".confidential_vm_encryption_type")
            security_profile.set_prop("secureVMDiskEncryptionSetId", AAZStrType, ".secure_vm_disk_encryption_set_id")

    @classmethod
    def _build_schema_gallery_disk_image_source_update(cls, _builder):
        if _builder is None:
            return
        _builder.set_prop("id", AAZStrType, ".id")
        _builder.set_prop("storageAccountId", AAZStrType, ".storage_account_id")
        _builder.set_prop("uri", AAZStrType, ".uri")

    @classmethod
    def _build_schema_uefi_key_update(cls, _builder):
        if _builder is None:
            return
        _builder.set_prop("type", AAZStrType, ".type")
        _builder.set_prop("value", AAZListType, ".value")

        value = _builder.get(".value")
        if value is not None:
            value.set_elements(AAZStrType, ".")

    _schema_encryption_images_read = None

    @classmethod
    def _build_schema_encryption_images_read(cls, _schema):
        if cls._schema_encryption_images_read is not None:
            _schema.data_disk_images = cls._schema_encryption_images_read.data_disk_images
            _schema.os_disk_image = cls._schema_encryption_images_read.os_disk_image
            return

        cls._schema_encryption_images_read = _schema_encryption_images_read = AAZObjectType()

        encryption_images_read = _schema_encryption_images_read
        encryption_images_read.data_disk_images = AAZListType(
            serialized_name="dataDiskImages",
        )
        encryption_images_read.os_disk_image = AAZObjectType(
            serialized_name="osDiskImage",
        )

        data_disk_images = _schema_encryption_images_read.data_disk_images
        data_disk_images.Element = AAZObjectType()

        _element = _schema_encryption_images_read.data_disk_images.Element
        _element.disk_encryption_set_id = AAZStrType(
            serialized_name="diskEncryptionSetId",
        )
        _element.lun = AAZIntType(
            flags={"required": True},
        )

        os_disk_image = _schema_encryption_images_read.os_disk_image
        os_disk_image.disk_encryption_set_id = AAZStrType(
            serialized_name="diskEncryptionSetId",
        )
        os_disk_image.security_profile = AAZObjectType(
            serialized_name="securityProfile",
        )

        security_profile = _schema_encryption_images_read.os_disk_image.security_profile
        security_profile.confidential_vm_encryption_type = AAZStrType(
            serialized_name="confidentialVMEncryptionType",
        )
        security_profile.secure_vm_disk_encryption_set_id = AAZStrType(
            serialized_name="secureVMDiskEncryptionSetId",
        )

        _schema.data_disk_images = cls._schema_encryption_images_read.data_disk_images
        _schema.os_disk_image = cls._schema_encryption_images_read.os_disk_image

    _schema_gallery_disk_image_source_read = None

    @classmethod
    def _build_schema_gallery_disk_image_source_read(cls, _schema):
        if cls._schema_gallery_disk_image_source_read is not None:
            _schema.id = cls._schema_gallery_disk_image_source_read.id
            _schema.storage_account_id = cls._schema_gallery_disk_image_source_read.storage_account_id
            _schema.uri = cls._schema_gallery_disk_image_source_read.uri
            return

        cls._schema_gallery_disk_image_source_read = _schema_gallery_disk_image_source_read = AAZObjectType()

        gallery_disk_image_source_read = _schema_gallery_disk_image_source_read
        gallery_disk_image_source_read.id = AAZStrType()
        gallery_disk_image_source_read.storage_account_id = AAZStrType(
            serialized_name="storageAccountId",
        )
        gallery_disk_image_source_read.uri = AAZStrType()

        _schema.id = cls._schema_gallery_disk_image_source_read.id
        _schema.storage_account_id = cls._schema_gallery_disk_image_source_read.storage_account_id
        _schema.uri = cls._schema_gallery_disk_image_source_read.uri

    _schema_gallery_image_version_read = None

    @classmethod
    def _build_schema_gallery_image_version_read(cls, _schema):
        if cls._schema_gallery_image_version_read is not None:
            _schema.id = cls._schema_gallery_image_version_read.id
            _schema.location = cls._schema_gallery_image_version_read.location
            _schema.name = cls._schema_gallery_image_version_read.name
            _schema.properties = cls._schema_gallery_image_version_read.properties
            _schema.tags = cls._schema_gallery_image_version_read.tags
            _schema.type = cls._schema_gallery_image_version_read.type
            return

        cls._schema_gallery_image_version_read = _schema_gallery_image_version_read = AAZObjectType()

        gallery_image_version_read = _schema_gallery_image_version_read
        gallery_image_version_read.id = AAZStrType(
            flags={"read_only": True},
        )
        gallery_image_version_read.location = AAZStrType(
            flags={"required": True},
        )
        gallery_image_version_read.name = AAZStrType(
            flags={"read_only": True},
        )
        gallery_image_version_read.properties = AAZObjectType(
            flags={"client_flatten": True},
        )
        gallery_image_version_read.tags = AAZDictType()
        gallery_image_version_read.type = AAZStrType(
            flags={"read_only": True},
        )

        properties = _schema_gallery_image_version_read.properties
        properties.provisioning_state = AAZStrType(
            serialized_name="provisioningState",
            flags={"read_only": True},
        )
        properties.publishing_profile = AAZObjectType(
            serialized_name="publishingProfile",
        )
        properties.replication_status = AAZObjectType(
            serialized_name="replicationStatus",
            flags={"read_only": True},
        )
        properties.restore = AAZBoolType()
        properties.safety_profile = AAZObjectType(
            serialized_name="safetyProfile",
        )
        properties.security_profile = AAZObjectType(
            serialized_name="securityProfile",
        )
        properties.storage_profile = AAZObjectType(
            serialized_name="storageProfile",
            flags={"required": True},
        )
        properties.validations_profile = AAZObjectType(
            serialized_name="validationsProfile",
            flags={"read_only": True},
        )

        publishing_profile = _schema_gallery_image_version_read.properties.publishing_profile
        publishing_profile.end_of_life_date = AAZStrType(
            serialized_name="endOfLifeDate",
        )
        publishing_profile.exclude_from_latest = AAZBoolType(
            serialized_name="excludeFromLatest",
        )
        publishing_profile.published_date = AAZStrType(
            serialized_name="publishedDate",
            flags={"read_only": True},
        )
        publishing_profile.replica_count = AAZIntType(
            serialized_name="replicaCount",
        )
        publishing_profile.replication_mode = AAZStrType(
            serialized_name="replicationMode",
        )
        publishing_profile.storage_account_type = AAZStrType(
            serialized_name="storageAccountType",
        )
        publishing_profile.target_extended_locations = AAZListType(
            serialized_name="targetExtendedLocations",
        )
        publishing_profile.target_regions = AAZListType(
            serialized_name="targetRegions",
        )

        target_extended_locations = _schema_gallery_image_version_read.properties.publishing_profile.target_extended_locations
        target_extended_locations.Element = AAZObjectType()

        _element = _schema_gallery_image_version_read.properties.publishing_profile.target_extended_locations.Element
        _element.encryption = AAZObjectType()
        cls._build_schema_encryption_images_read(_element.encryption)
        _element.extended_location = AAZObjectType(
            serialized_name="extendedLocation",
        )
        _element.extended_location_replica_count = AAZIntType(
            serialized_name="extendedLocationReplicaCount",
        )
        _element.name = AAZStrType()
        _element.storage_account_type = AAZStrType(
            serialized_name="storageAccountType",
        )

        extended_location = _schema_gallery_image_version_read.properties.publishing_profile.target_extended_locations.Element.extended_location
        extended_location.name = AAZStrType()
        extended_location.type = AAZStrType()

        target_regions = _schema_gallery_image_version_read.properties.publishing_profile.target_regions
        target_regions.Element = AAZObjectType()

        _element = _schema_gallery_image_version_read.properties.publishing_profile.target_regions.Element
        _element.additional_replica_sets = AAZListType(
            serialized_name="additionalReplicaSets",
        )
        _element.encryption = AAZObjectType()
        cls._build_schema_encryption_images_read(_element.encryption)
        _element.exclude_from_latest = AAZBoolType(
            serialized_name="excludeFromLatest",
        )
        _element.name = AAZStrType(
            flags={"required": True},
        )
        _element.regional_replica_count = AAZIntType(
            serialized_name="regionalReplicaCount",
        )
        _element.storage_account_type = AAZStrType(
            serialized_name="storageAccountType",
        )

        additional_replica_sets = _schema_gallery_image_version_read.properties.publishing_profile.target_regions.Element.additional_replica_sets
        additional_replica_sets.Element = AAZObjectType()

        _element = _schema_gallery_image_version_read.properties.publishing_profile.target_regions.Element.additional_replica_sets.Element
        _element.regional_replica_count = AAZIntType(
            serialized_name="regionalReplicaCount",
        )
        _element.storage_account_type = AAZStrType(
            serialized_name="storageAccountType",
        )

        replication_status = _schema_gallery_image_version_read.properties.replication_status
        replication_status.aggregated_state = AAZStrType(
            serialized_name="aggregatedState",
            flags={"read_only": True},
        )
        replication_status.summary = AAZListType(
            flags={"read_only": True},
        )

        summary = _schema_gallery_image_version_read.properties.replication_status.summary
        summary.Element = AAZObjectType()

        _element = _schema_gallery_image_version_read.properties.replication_status.summary.Element
        _element.details = AAZStrType(
            flags={"read_only": True},
        )
        _element.progress = AAZIntType(
            flags={"read_only": True},
        )
        _element.region = AAZStrType(
            flags={"read_only": True},
        )
        _element.state = AAZStrType(
            flags={"read_only": True},
        )

        safety_profile = _schema_gallery_image_version_read.properties.safety_profile
        safety_profile.allow_deletion_of_replicated_locations = AAZBoolType(
            serialized_name="allowDeletionOfReplicatedLocations",
        )
        safety_profile.block_deletion_before_end_of_life = AAZBoolType(
            serialized_name="blockDeletionBeforeEndOfLife",
        )
        safety_profile.policy_violations = AAZListType(
            serialized_name="policyViolations",
            flags={"read_only": True},
        )
        safety_profile.reported_for_policy_violation = AAZBoolType(
            serialized_name="reportedForPolicyViolation",
            flags={"read_only": True},
        )

        policy_violations = _schema_gallery_image_version_read.properties.safety_profile.policy_violations
        policy_violations.Element = AAZObjectType()

        _element = _schema_gallery_image_version_read.properties.safety_profile.policy_violations.Element
        _element.category = AAZStrType()
        _element.details = AAZStrType()

        security_profile = _schema_gallery_image_version_read.properties.security_profile
        security_profile.uefi_settings = AAZObjectType(
            serialized_name="uefiSettings",
        )

        uefi_settings = _schema_gallery_image_version_read.properties.security_profile.uefi_settings
        uefi_settings.additional_signatures = AAZObjectType(
            serialized_name="additionalSignatures",
        )
        uefi_settings.signature_template_names = AAZListType(
            serialized_name="signatureTemplateNames",
        )

        additional_signatures = _schema_gallery_image_version_read.properties.security_profile.uefi_settings.additional_signatures
        additional_signatures.db = AAZListType()
        additional_signatures.dbx = AAZListType()
        additional_signatures.kek = AAZListType()
        additional_signatures.pk = AAZObjectType()
        cls._build_schema_uefi_key_read(additional_signatures.pk)

        db = _schema_gallery_image_version_read.properties.security_profile.uefi_settings.additional_signatures.db
        db.Element = AAZObjectType()
        cls._build_schema_uefi_key_read(db.Element)

        dbx = _schema_gallery_image_version_read.properties.security_profile.uefi_settings.additional_signatures.dbx
        dbx.Element = AAZObjectType()
        cls._build_schema_uefi_key_read(dbx.Element)

        kek = _schema_gallery_image_version_read.properties.security_profile.uefi_settings.additional_signatures.kek
        kek.Element = AAZObjectType()
        cls._build_schema_uefi_key_read(kek.Element)

        signature_template_names = _schema_gallery_image_version_read.properties.security_profile.uefi_settings.signature_template_names
        signature_template_names.Element = AAZStrType()

        storage_profile = _schema_gallery_image_version_read.properties.storage_profile
        storage_profile.data_disk_images = AAZListType(
            serialized_name="dataDiskImages",
        )
        storage_profile.os_disk_image = AAZObjectType(
            serialized_name="osDiskImage",
        )
        storage_profile.source = AAZObjectType()

        data_disk_images = _schema_gallery_image_version_read.properties.storage_profile.data_disk_images
        data_disk_images.Element = AAZObjectType()

        _element = _schema_gallery_image_version_read.properties.storage_profile.data_disk_images.Element
        _element.host_caching = AAZStrType(
            serialized_name="hostCaching",
        )
        _element.lun = AAZIntType(
            flags={"required": True},
        )
        _element.size_in_gb = AAZIntType(
            serialized_name="sizeInGB",
            flags={"read_only": True},
        )
        _element.source = AAZObjectType()
        cls._build_schema_gallery_disk_image_source_read(_element.source)

        os_disk_image = _schema_gallery_image_version_read.properties.storage_profile.os_disk_image
        os_disk_image.host_caching = AAZStrType(
            serialized_name="hostCaching",
        )
        os_disk_image.size_in_gb = AAZIntType(
            serialized_name="sizeInGB",
            flags={"read_only": True},
        )
        os_disk_image.source = AAZObjectType()
        cls._build_schema_gallery_disk_image_source_read(os_disk_image.source)

        source = _schema_gallery_image_version_read.properties.storage_profile.source
        source.community_gallery_image_id = AAZStrType(
            serialized_name="communityGalleryImageId",
        )
        source.id = AAZStrType()
        source.virtual_machine_id = AAZStrType(
            serialized_name="virtualMachineId",
        )

        validations_profile = _schema_gallery_image_version_read.properties.validations_profile
        validations_profile.executed_validations = AAZListType(
            serialized_name="executedValidations",
        )
        validations_profile.platform_attributes = AAZListType(
            serialized_name="platformAttributes",
        )
        validations_profile.validation_etag = AAZStrType(
            serialized_name="validationEtag",
        )

        executed_validations = _schema_gallery_image_version_read.properties.validations_profile.executed_validations
        executed_validations.Element = AAZObjectType()

        _element = _schema_gallery_image_version_read.properties.validations_profile.executed_validations.Element
        _element.execution_time = AAZStrType(
            serialized_name="executionTime",
        )
        _element.status = AAZStrType()
        _element.type = AAZStrType()
        _element.version = AAZStrType()

        platform_attributes = _schema_gallery_image_version_read.properties.validations_profile.platform_attributes
        platform_attributes.Element = AAZObjectType()

        _element = _schema_gallery_image_version_read.properties.validations_profile.platform_attributes.Element
        _element.name = AAZStrType(
            flags={"read_only": True},
        )
        _element.value = AAZStrType(
            flags={"read_only": True},
        )

        tags = _schema_gallery_image_version_read.tags
        tags.Element = AAZStrType()

        _schema.id = cls._schema_gallery_image_version_read.id
        _schema.location = cls._schema_gallery_image_version_read.location
        _schema.name = cls._schema_gallery_image_version_read.name
        _schema.properties = cls._schema_gallery_image_version_read.properties
        _schema.tags = cls._schema_gallery_image_version_read.tags
        _schema.type = cls._schema_gallery_image_version_read.type

    _schema_uefi_key_read = None

    @classmethod
    def _build_schema_uefi_key_read(cls, _schema):
        if cls._schema_uefi_key_read is not None:
            _schema.type = cls._schema_uefi_key_read.type
            _schema.value = cls._schema_uefi_key_read.value
            return

        cls._schema_uefi_key_read = _schema_uefi_key_read = AAZObjectType()

        uefi_key_read = _schema_uefi_key_read
        uefi_key_read.type = AAZStrType()
        uefi_key_read.value = AAZListType()

        value = _schema_uefi_key_read.value
        value.Element = AAZStrType()

        _schema.type = cls._schema_uefi_key_read.type
        _schema.value = cls._schema_uefi_key_read.value


__all__ = ["Update"]



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
#
# Code generated by aaz-dev-tools
# --------------------------------------------------------------------------------------------

# pylint: skip-file
# flake8: noqa

from azure.cli.core.aaz import *


@register_command(
    "lab environment create",
    is_preview=True,
)
class Create(AAZCommand):
    """Create an existing environment. This operation can take a while to complete.
    """

    _aaz_info = {
        "version": "2018-09-15",
        "resources": [
            ["mgmt-plane", "/subscriptions/{}/resourcegroups/{}/providers/microsoft.devtestlab/labs/{}/users/{}/environments/{}", "2018-09-15"],
        ]
    }

    AZ_SUPPORT_NO_WAIT = True

    def _handler(self, command_args):
        super()._handler(command_args)
        return self.build_lro_poller(self._execute_operations, self._output)

    _args_schema = None

    @classmethod
    def _build_arguments_schema(cls, *args, **kwargs):
        if cls._args_schema is not None:
            return cls._args_schema
        cls._args_schema = super()._build_arguments_schema(*args, **kwargs)

        # define Arg Group ""

        _args_schema = cls._args_schema
        _args_schema.lab_name = AAZStrArg(
            options=["--lab-name"],
            help="The name of the lab.",
            required=True,
        )
        _args_schema.name = AAZStrArg(
            options=["-n", "--name"],
            help="The name of the environment.",
            required=True,
        )
        _args_schema.resource_group = AAZResourceGroupNameArg(
            required=True,
        )
        _args_schema.user_name = AAZStrArg(
            options=["--user-name"],
            help="The name of the user profile.",
            required=True,
        )
        _args_schema.tags = AAZDictArg(
            options=["--tags"],
            help="The tags of the resource.",
        )

        tags = cls._args_schema.tags
        tags.Element = AAZStrArg()

        # define Arg Group "DtlEnvironment"

        # define Arg Group "Properties"

        _args_schema = cls._args_schema
        _args_schema.deployment_properties = AAZObjectArg(
            options=["--deployment-properties"],
            arg_group="Properties",
            help="The deployment properties of the environment.",
        )

        deployment_properties = cls._args_schema.deployment_properties
        deployment_properties.arm_template_id = AAZStrArg(
            options=["arm-template-id"],
            help="The Azure Resource Manager template's identifier.",
        )
        deployment_properties.parameters = AAZListArg(
            options=["parameters"],
            help="The parameters of the Azure Resource Manager template.",
        )

        parameters = cls._args_schema.deployment_properties.parameters
        parameters.Element = AAZObjectArg()

        _element = cls._args_schema.deployment_properties.parameters.Element
        _element.name = AAZStrArg(
            options=["name"],
            help="The name of the template parameter.",
        )
        _element.value = AAZStrArg(
            options=["value"],
            help="The value of the template parameter.",
        )
        return cls._args_schema

    def _execute_operations(self):
        self.pre_operations()
        yield self.EnvironmentsCreateOrUpdate(ctx=self.ctx)()
        self.post_operations()

    @register_callback
    def pre_operations(self):
        pass

    @register_callback
    def post_operations(self):
        pass

    def _output(self, *args, **kwargs):
        result = self.deserialize_output(self.ctx.vars.instance, client_flatten=True)
        return result

    class EnvironmentsCreateOrUpdate(AAZHttpOperation):
        CLIENT_TYPE = "MgmtClient"

        def __call__(self, *args, **kwargs):
            request = self.make_request()
            session = self.client.send_request(request=request, stream=False, **kwargs)
            if session.http_response.status_code in [202]:
                return self.client.build_lro_polling(
                    self.ctx.args.no_wait,
                    session,
                    self.on_200_201,
                    self.on_error,
                    lro_options={"final-state-via": "azure-async-operation"},
                    path_format_arguments=self.url_parameters,
                )
            if session.http_response.status_code in [200, 201]:
                return self.client.build_lro_polling(
                    self.ctx.args.no_wait,
                    session,
                    self.on_200_201,
                    self.on_error,
                    lro_options={"final-state-via": "azure-async-operation"},
                    path_format_arguments=self.url_parameters,
                )

            return self.on_error(session.http_response)

        @property
        def url(self):
            return self.client.format_url(
                "/subscriptions/{subscriptionId}/resourceGroups/{resourceGroupName}/providers/Microsoft.DevTestLab/labs/{labName}/users/{userName}/environments/{name}",
                **self.url_parameters
            )

        @property
        def method(self):
            return "PUT"

        @property
        def error_format(self):
            return "ODataV4Format"

        @property
        def url_parameters(self):
            parameters = {
                **self.serialize_url_param(
                    "labName", self.ctx.args.lab_name,
                    required=True,
                ),
                **self.serialize_url_param(
                    "name", self.ctx.args.name,
                    required=True,
                ),
                **self.serialize_url_param(
                    "resourceGroupName", self.ctx.args.resource_group,
                    required=True,
                ),
                **self.serialize_url_param(
                    "subscriptionId", self.ctx.subscription_id,
                    required=True,
                ),
                **self.serialize_url_param(
                    "userName", self.ctx.args.user_name,
                    required=True,
                ),
            }
            return parameters

        @property
        def query_parameters(self):
            parameters = {
                **self.serialize_query_param(
                    "api-version", "2018-09-15",
                    required=True,
                ),
            }
            return parameters

        @property
        def header_parameters(self):
            parameters = {
                **self.serialize_header_param(
                    "Content-Type", "application/json",
                ),
                **self.serialize_header_param(
                    "Accept", "application/json",
                ),
            }
            return parameters

        @property
        def content(self):
            _content_value, _builder = self.new_content_builder(
                self.ctx.args,
                typ=AAZObjectType,
                typ_kwargs={"flags": {"required": True, "client_flatten": True}}
            )
            _builder.set_prop("properties", AAZObjectType, ".", typ_kwargs={"flags": {"required": True, "client_flatten": True}})
            _builder.set_prop("tags", AAZDictType, ".tags")

            properties = _builder.get(".properties")
            if properties is not None:
                properties.set_prop("deploymentProperties", AAZObjectType, ".deployment_properties")

            deployment_properties = _builder.get(".properties.deploymentProperties")
            if deployment_properties is not None:
                deployment_properties.set_prop("armTemplateId", AAZStrType, ".arm_template_id")
                deployment_properties.set_prop("parameters", AAZListType, ".parameters")

            parameters = _builder.get(".properties.deploymentProperties.parameters")
            if parameters is not None:
                parameters.set_elements(AAZObjectType, ".")

            _elements = _builder.get(".properties.deploymentProperties.parameters[]")
            if _elements is not None:
                _elements.set_prop("name", AAZStrType, ".name")
                _elements.set_prop("value", AAZStrType, ".value")

            tags = _builder.get(".tags")
            if tags is not None:
                tags.set_elements(AAZStrType, ".")

            return self.serialize_content(_content_value)

        def on_200_201(self, session):
            data = self.deserialize_http_content(session)
            self.ctx.set_var(
                "instance",
                data,
                schema_builder=self._build_schema_on_200_201
            )

        _schema_on_200_201 = None

        @classmethod
        def _build_schema_on_200_201(cls):
            if cls._schema_on_200_201 is not None:
                return cls._schema_on_200_201

            cls._schema_on_200_201 = AAZObjectType()

            _schema_on_200_201 = cls._schema_on_200_201
            _schema_on_200_201.id = AAZStrType(
                flags={"read_only": True},
            )
            _schema_on_200_201.location = AAZStrType()
            _schema_on_200_201.name = AAZStrType(
                flags={"read_only": True},
            )
            _schema_on_200_201.properties = AAZObjectType(
                flags={"required": True, "client_flatten": True},
            )
            _schema_on_200_201.tags = AAZDictType()
            _schema_on_200_201.type = AAZStrType(
                flags={"read_only": True},
            )

            properties = cls._schema_on_200_201.properties
            properties.arm_template_display_name = AAZStrType(
                serialized_name="armTemplateDisplayName",
            )
            properties.created_by_user = AAZStrType(
                serialized_name="createdByUser",
                flags={"read_only": True},
            )
            properties.deployment_properties = AAZObjectType(
                serialized_name="deploymentProperties",
            )
            properties.provisioning_state = AAZStrType(
                serialized_name="provisioningState",
                flags={"read_only": True},
            )
            properties.resource_group_id = AAZStrType(
                serialized_name="resourceGroupId",
                flags={"read_only": True},
            )
            properties.unique_identifier = AAZStrType(
                serialized_name="uniqueIdentifier",
                flags={"read_only": True},
            )

            deployment_properties = cls._schema_on_200_201.properties.deployment_properties
            deployment_properties.arm_template_id = AAZStrType(
                serialized_name="armTemplateId",
            )
            deployment_properties.parameters = AAZListType()

            parameters = cls._schema_on_200_201.properties.deployment_properties.parameters
            parameters.Element = AAZObjectType()

            _element = cls._schema_on_200_201.properties.deployment_properties.parameters.Element
            _element.name = AAZStrType()
            _element.value = AAZStrType()

            tags = cls._schema_on_200_201.tags
            tags.Element = AAZStrType()

            return cls._schema_on_200_201


class _CreateHelper:
    """Helper class for Create"""


__all__ = ["Create"]



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

from itertools import groupby

from azure.mgmt.resource.deployments.models import ChangeType, PropertyChangeType, Level

from ._symbol import Symbol
from ._color import Color, ColoredStringBuilder
from ._utils import split_resource_id

_change_type_to_color = {
    ChangeType.create: Color.GREEN,
    ChangeType.delete: Color.ORANGE,
    ChangeType.modify: Color.PURPLE,
    ChangeType.deploy: Color.BLUE,
    ChangeType.no_change: Color.RESET,
    ChangeType.ignore: Color.GRAY,
    ChangeType.unsupported: Color.GRAY,
    PropertyChangeType.no_effect: Color.GRAY,
}

_property_change_type_to_color = {
    PropertyChangeType.create: Color.GREEN,
    PropertyChangeType.delete: Color.ORANGE,
    PropertyChangeType.modify: Color.PURPLE,
    PropertyChangeType.array: Color.PURPLE,
    PropertyChangeType.no_effect: Color.GRAY,
}

_diagnostic_level_to_color = {
    Level.ERROR: Color.RED,
    Level.WARNING: Color.DARK_YELLOW,
    Level.INFO: Color.RESET,
}

_change_type_to_symbol = {
    ChangeType.create: Symbol.PLUS,
    ChangeType.delete: Symbol.MINUS,
    ChangeType.modify: Symbol.TILDE,
    ChangeType.deploy: Symbol.EXCLAMATION_POINT,
    ChangeType.no_change: Symbol.EQUAL,
    ChangeType.ignore: Symbol.ASTERISK,
    ChangeType.unsupported: Symbol.CROSS,
    PropertyChangeType.no_effect: Symbol.CROSS,
}

_property_change_type_to_symbol = {
    PropertyChangeType.create: Symbol.PLUS,
    PropertyChangeType.delete: Symbol.MINUS,
    PropertyChangeType.modify: Symbol.TILDE,
    PropertyChangeType.array: Symbol.TILDE,
    PropertyChangeType.no_effect: Symbol.CROSS,
}

_change_type_to_weight = {
    ChangeType.delete: 0,
    ChangeType.create: 1,
    ChangeType.deploy: 2,
    ChangeType.modify: 3,
    ChangeType.no_change: 4,
    ChangeType.unsupported: 5,
    ChangeType.ignore: 6,
    PropertyChangeType.no_effect: 7,
}

_property_change_type_to_weight = {
    PropertyChangeType.delete: 0,
    PropertyChangeType.create: 1,
    PropertyChangeType.modify: 2,
    PropertyChangeType.array: 2,
    PropertyChangeType.no_effect: 3,
}


def format_what_if_operation_result(what_if_operation_result, enable_color=True):
    builder = ColoredStringBuilder(enable_color)
    _format_noise_notice(builder)
    _format_change_type_legend(builder, what_if_operation_result.changes)
    _format_resource_changes(builder,
                             what_if_operation_result.changes,
                             definite_changes=True)
    _format_resource_changes_stats(builder,
                                   what_if_operation_result.changes,
                                   definite_changes=True)
    _format_resource_changes(builder,
                             what_if_operation_result.potential_changes,
                             definite_changes=False)
    _format_resource_changes_stats(builder,
                                   what_if_operation_result.potential_changes,
                                   definite_changes=False)
    _format_diagnostics(builder,
                        what_if_operation_result.changes,
                        what_if_operation_result.potential_changes,
                        what_if_operation_result.diagnostics)
    return builder.build()


def _format_noise_notice(builder):
    builder.append_line(
        """Note: The result may contain false positive predictions (noise).
You can help us improve the accuracy of the result by opening an issue here: https://aka.ms/WhatIfIssues"""
    )
    builder.append_line()


def _format_change_type_legend(builder, resource_changes):
    if not resource_changes:
        return

    def populate_change_type_set(property_changes):
        if not property_changes:
            return

        for property_change in property_changes:
            property_change_type = property_change.property_change_type
            change_type_set.add(
                ChangeType.modify if property_change_type == PropertyChangeType.array else property_change_type
            )
            populate_change_type_set(property_change.children)

    change_type_set = set()

    for resource_change in resource_changes:
        change_type_set.add(resource_change.change_type)
        populate_change_type_set(resource_change.delta)

    change_types = sorted(change_type_set, key=lambda x: _change_type_to_weight[x])

    builder.append("Resource and property changes are indicated with ")
    builder.append_line("this symbol:" if len(change_types) == 1 else "these symbols:")

    for change_type in change_types:
        change_type_symbol = _change_type_to_symbol[change_type]
        change_type_color = _change_type_to_color[change_type]
        _format_indent(builder)
        builder.append(change_type_symbol, change_type_color).append(Symbol.WHITE_SPACE)
        builder.append_line(change_type.title())


def _format_resource_changes_stats(builder, resource_changes, definite_changes=True):
    if definite_changes:
        builder.append_line().append("Resource changes: ")

        if not resource_changes:
            builder.append("no change.")
            return
    elif resource_changes:
        builder.append_line().append("Potential changes: ")

    if resource_changes:
        sorted_resource_changes = sorted(resource_changes, key=lambda x: _change_type_to_weight[x.change_type])
        resource_changes_by_change_type = groupby(sorted_resource_changes, lambda x: x.change_type)
        count_by_change_type = map(lambda x: (x[0], len(list(x[1]))), resource_changes_by_change_type)
        count_by_change_type = filter(lambda x: x[1] > 0, count_by_change_type)
        change_type_stats = map(lambda x: _format_change_type_count(x[0], x[1]), count_by_change_type)

        builder.append(", ".join(change_type_stats)).append(".")


def _format_change_type_count(change_type, count):  # pylint: disable=too-many-return-statements
    if change_type == ChangeType.create:
        return f"{count} to create"
    if change_type == ChangeType.delete:
        return f"{count} to delete"
    if change_type == ChangeType.deploy:
        return f"{count} to deploy"
    if change_type == ChangeType.modify:
        return f"{count} to modify"
    if change_type == ChangeType.ignore:
        return f"{count} to ignore"
    if change_type == ChangeType.no_change:
        return f"{count} no change"
    if change_type == ChangeType.unsupported:
        return f"{count} unsupported"

    raise ValueError(f"Invalid ChangeType: {change_type}")


def _format_diagnostics(builder, resource_changes, potential_changes, diagnostics):
    short_circuited_resources = [r for r in (resource_changes or []) + (potential_changes or [])
                                 if r.change_type == ChangeType.UNSUPPORTED]

    diags = diagnostics or []

    if len(short_circuited_resources) + len(diags) > 0:
        builder.append_line()
        builder.append_line()
        builder.append(f"Diagnostics ({len(short_circuited_resources) + len(diags)}): ")
        builder.append_line()

        for change in short_circuited_resources:
            with builder.new_color_scope(_diagnostic_level_to_color[Level.WARNING]):
                builder.append(change.resource_id)
                builder.append(" (Unsupported) ")
                builder.append(change.unsupported_reason)
                builder.append_line()

        for diag in diags:
            with builder.new_color_scope(_diagnostic_level_to_color[diag.level]):
                builder.append(diag.target)
                builder.append(" (")
                builder.append(diag.code)
                builder.append(") ")
                builder.append(diag.message)
                builder.append_line()


def _format_resource_changes(builder, resource_changes, definite_changes=True):
    if not resource_changes:
        return

    num_scopes = len(set(map(_get_scope_uppercase, resource_changes)))
    resource_changes_by_scope = groupby(sorted(resource_changes, key=_get_scope_uppercase), _get_scope_uppercase)

    builder.append_line()
    if definite_changes:
        builder.append("The deployment will update the following ")
    else:
        builder.append("The following change MAY OR MAY NOT be deployed to the following ")

    builder.append_line('scope:' if num_scopes == 1 else 'scopes:')

    for _, resource_changes_in_scope in resource_changes_by_scope:
        resource_changes_in_scope_list = list(resource_changes_in_scope)
        scope = _get_scope(resource_changes_in_scope_list[0])
        _format_resource_changes_in_scope(builder, scope, resource_changes_in_scope_list)


def _format_resource_changes_in_scope(builder, scope, resource_changes_in_scope):
    builder.append_line().append_line(f"Scope: {scope}")

    sorted_resource_changes = sorted(
        resource_changes_in_scope,
        # Sort by change_type and then by relative_resource_id.
        key=lambda x: (_change_type_to_weight[x.change_type], _get_relative_resource_id(x)),
    )

    for change_type, resource_changes in groupby(sorted_resource_changes, lambda x: x.change_type):
        with builder.new_color_scope(_change_type_to_color[change_type]):
            for resource_change in resource_changes:
                is_last = resource_change == sorted_resource_changes[-1]
                _format_resource_change(builder, resource_change, is_last)


def _format_resource_change(builder, resource_change, is_last):
    change_type = resource_change.change_type
    relative_resource_id = _get_relative_resource_id(resource_change)
    api_version = _get_api_version(resource_change)

    builder.append_line()
    _format_resource_change_path(builder, change_type, relative_resource_id, api_version)

    if change_type == ChangeType.create and resource_change.after:
        _format_json(builder, resource_change.after, indent_level=2)

    elif change_type == ChangeType.delete and resource_change.before:
        _format_json(builder, resource_change.before, indent_level=2)

    elif resource_change.delta:
        with builder.new_color_scope(Color.RESET):
            builder.append_line()
            _format_property_changes(
                builder,
                sorted(
                    resource_change.delta,
                    key=lambda x: (_property_change_type_to_weight[x.property_change_type], x.path),
                ),
            )

    elif is_last:
        builder.append_line()


def _format_resource_change_path(builder, change_type, resource_change_id, api_version):
    return _format_path(
        builder,
        resource_change_id,
        0,
        1,
        lambda builder: _format_resource_change_type(builder, change_type),
        lambda builder: _format_resource_change_api_verion(builder, api_version),
    )


def _format_resource_change_type(builder, change_type):
    change_symbol = _change_type_to_symbol[change_type]
    builder.append(change_symbol).append(Symbol.WHITE_SPACE)


def _format_resource_change_api_verion(builder, api_version):
    if not api_version:
        return

    with builder.new_color_scope(Color.RESET):
        builder.append(Symbol.WHITE_SPACE)
        builder.append(Symbol.LEFT_SQUARE_BRACKET)
        builder.append(api_version)
        builder.append(Symbol.RIGHT_SQUARE_BRACKET)


def _format_property_changes(builder, property_changes, indent_level=2):
    max_path_length = _get_max_path_length_from_property_changes(property_changes)

    for property_change in property_changes:
        _format_property_change(builder, property_change, max_path_length, indent_level)
        builder.append_line()


def _format_property_change(builder, property_change, max_path_length, indent_level):
    property_change_type = property_change.property_change_type
    before, after, children = property_change.before, property_change.after, property_change.children

    if property_change_type == PropertyChangeType.create:
        _format_property_change_path(builder, property_change, "after", max_path_length, indent_level)
        _format_property_create(builder, after, indent_level + 1)
    elif property_change_type == PropertyChangeType.delete:
        _format_property_change_path(builder, property_change, "before", max_path_length, indent_level)
        _format_property_delete(builder, before, indent_level + 1)
    elif property_change_type == PropertyChangeType.modify:
        _format_property_change_path(builder, property_change, "before", max_path_length, indent_level)
        _format_property_modify(builder, before, after, children, indent_level + 1)
    elif property_change_type == PropertyChangeType.array:
        _format_property_change_path(builder, property_change, "children", max_path_length, indent_level)
        _format_property_array_change(builder, property_change, children, indent_level + 1)
    elif property_change_type == PropertyChangeType.no_effect:
        _format_property_change_path(builder, property_change, "after", max_path_length, indent_level)
        _format_property_no_effect(builder, after, indent_level + 1)
    else:
        raise ValueError(f"Unknown property change type: {property_change_type}.")


def _format_property_change_path(builder, property_change, value_name, max_path_length, indent_level):
    if not property_change.path:
        return

    path, property_change_type = property_change.path, property_change.property_change_type
    value = getattr(property_change, value_name)

    padding_width = max_path_length - len(path) + 1

    if _is_non_empty_array(value):
        padding_width = 1
    if _is_non_empty_object(value):
        padding_width = 0
    if property_change_type == PropertyChangeType.modify and property_change.children:
        # Has nested changes.
        padding_width = 0

    _format_path(
        builder,
        path,
        padding_width,
        indent_level,
        format_head=lambda builder: _format_property_change_type(builder, property_change_type),
        format_tail=_format_colon,
    )


def _format_property_change_type(builder, property_change_type):
    property_change_symbol = _property_change_type_to_symbol[property_change_type]
    property_change_color = _property_change_type_to_color[property_change_type]
    builder.append(property_change_symbol, property_change_color).append(Symbol.WHITE_SPACE)


def _format_property_no_effect(builder, value, indent_level):
    with builder.new_color_scope(_property_change_type_to_color[PropertyChangeType.no_effect]):
        _format_json(builder, value, indent_level=indent_level)


def _format_property_create(builder, value, indent_level):
    with builder.new_color_scope(_property_change_type_to_color[PropertyChangeType.create]):
        _format_json(builder, value, indent_level=indent_level)


def _format_property_delete(builder, value, indent_level):
    with builder.new_color_scope(_property_change_type_to_color[PropertyChangeType.delete]):
        _format_json(builder, value, indent_level=indent_level)


def _format_property_modify(builder, before, after, children, indent_level):
    if children:
        # Has nested changes.
        builder.append_line().append_line()
        _format_property_changes(
            builder,
            sorted(children, key=lambda x: (_property_change_type_to_weight[x.property_change_type], x.path)),
            indent_level,
        )
    else:
        _format_property_delete(builder, before, indent_level)

        # Space before =>
        if _is_non_empty_object(before):
            builder.append_line()
            _format_indent(builder, indent_level)
        else:
            builder.append(Symbol.WHITE_SPACE)

        builder.append("=>")

        # Space after =>
        if not _is_non_empty_object(after):
            builder.append(Symbol.WHITE_SPACE)

        _format_property_create(builder, after, indent_level)

        if not _is_leaf(before) and _is_leaf(after):
            builder.append_line()


def _format_property_array_change(builder, parent_property_change, property_changes, indent_level):
    if not parent_property_change.path:
        # The parent change doesn't have a path, which means the current
        # array change is a nested change. We need to decrease indent_level
        # and print indentation before printing "[".
        indent_level -= 1
        _format_indent(builder, indent_level)

    if not property_changes:
        builder.append_line("[]")
        return

    # [
    builder.append(Symbol.LEFT_SQUARE_BRACKET).append_line()

    _format_property_changes(
        builder,
        sorted(property_changes, key=lambda x: (int(x.path), _property_change_type_to_weight[x.property_change_type])),
        indent_level,
    )

    # ]
    _format_indent(builder, indent_level)
    builder.append(Symbol.RIGHT_SQUARE_BRACKET)


def _get_api_version(resource_change):
    if resource_change.before:
        return resource_change.before.get("apiVersion")
    if resource_change.after:
        return resource_change.after.get("apiVersion")
    return None


def _get_scope(resource_change):
    scope, _ = split_resource_id(resource_change.resource_id)
    return scope


def _get_scope_uppercase(resource_change):
    return _get_scope(resource_change).upper()


def _get_relative_resource_id(resource_change):
    _, relative_resource_id = split_resource_id(resource_change.resource_id)
    return relative_resource_id


def _get_max_path_length_from_property_changes(property_changes):
    if not property_changes:
        return 0

    filtered_property_changes = filter(_should_consider_property_change_path, property_changes)
    path_lengths = map(lambda x: len(x.path), filtered_property_changes)

    return max(path_lengths, default=0)


def _should_consider_property_change_path(property_change):
    property_change_type = property_change.property_change_type

    if property_change_type == PropertyChangeType.create:
        return _is_leaf(property_change.after)

    if property_change_type in (PropertyChangeType.delete, PropertyChangeType.modify):
        return _is_leaf(property_change.before)

    return not property_change.children


def format_json(value, enable_color=True):
    builder = ColoredStringBuilder(enable_color)
    _format_json(builder, value)
    return builder.build()


def _format_json(builder, value, path="", max_path_length=0, indent_level=0):
    if _is_leaf(value):
        _format_json_path(builder, path, max_path_length - len(path) + 1, indent_level)
        _format_leaf(builder, value)
    elif _is_non_empty_array(value):
        _format_json_path(builder, path, 1, indent_level)
        _format_non_empty_array(builder, value, indent_level)
    elif _is_non_empty_object(value):
        _format_non_empty_object(builder, value, path, max_path_length, indent_level)
    else:
        raise ValueError(f"Invalid JSON value: {value}")


def _format_leaf(builder, value):
    if value is None:
        builder.append("null")
    elif isinstance(value, bool):
        builder.append(str(value).lower())
    elif isinstance(value, str):
        builder.append(Symbol.QUOTE).append(value).append(Symbol.QUOTE)
    else:
        builder.append(value)


def _format_non_empty_array(builder, value, indent_level):
    builder.append(Symbol.LEFT_SQUARE_BRACKET, Color.RESET).append_line()

    max_path_length = _get_max_path_length_from_array(value)

    for index, child_value in enumerate(value):
        child_path = str(index)

        if _is_non_empty_object(child_value):
            _format_json_path(builder, child_path, 0, indent_level + 1)
            _format_non_empty_object(builder, child_value, indent_level=indent_level + 1)
        else:
            _format_json(builder, child_value, child_path, max_path_length, indent_level + 1)

        builder.append_line()

    _format_indent(builder, indent_level)
    builder.append(Symbol.RIGHT_SQUARE_BRACKET, Color.RESET)


def _format_non_empty_object(builder, value, path=None, max_path_length=0, indent_level=0):
    is_root = not path

    if not path:
        # root object.
        builder.append_line().append_line()
        max_path_length = _get_max_path_length_from_object(value)
        indent_level += 1

    for child_path, child_value in value.items():
        child_path = child_path if is_root else f"{path}{Symbol.DOT}{child_path}"
        _format_json(builder, child_value, child_path, max_path_length, indent_level)

        # if _is_non_empty_array(child_value) or _is_leaf(child_value):
        if not _is_non_empty_object(child_value):
            builder.append_line()


def _format_json_path(builder, path, padding_width, indent_level):
    return _format_path(builder, path, padding_width, indent_level, format_tail=_format_colon)


def _format_path(builder, path, padding_width, indent_level, format_head=None, format_tail=None):
    if not path:
        return

    _format_indent(builder, indent_level)

    if format_head:
        format_head(builder)

    builder.append(path)

    if format_tail:
        format_tail(builder)

    builder.append(str(Symbol.WHITE_SPACE) * padding_width)


def _format_colon(builder):
    builder.append(Symbol.COLON, Color.RESET)


def _format_indent(builder, indent_level=1):
    builder.append(str(Symbol.WHITE_SPACE) * 2 * indent_level)


def _get_max_path_length_from_array(value):
    maxLengthIndex = 0

    for index, child_value in enumerate(value):
        if _is_leaf(child_value):
            maxLengthIndex = index

    return len(str(maxLengthIndex))


def _get_max_path_length_from_object(value):
    max_path_length = 0

    for key, child_value in value.items():
        if _is_non_empty_array(child_value):
            # Ignoring array paths to avoid long padding like this:
            #
            #   short.path:                   "foo"
            #   another.short.path:           "bar"
            #   very.very.long.path.to.array: [
            #     ...
            #   ]
            #   path.after.array:             "foobar"
            #
            # The following is what we want:
            #
            #   short.path:         "foo"
            #   another.short.path: "bar"
            #   very.very.long.path.to.array: [
            #     ...
            #   ]
            #   path.after.array:   "foobar"
            #
            continue

        current_path_length = (
            # Add one for dot.
            len(key) + 1 + _get_max_path_length_from_object(child_value)
            if _is_non_empty_object(child_value)
            else len(key)
        )

        max_path_length = max(max_path_length, current_path_length)

    return max_path_length


def _is_leaf(value):
    return (
        isinstance(value, (type(None), bool, int, float, str)) or
        (isinstance(value, list) and not value) or
        (isinstance(value, dict) and not value)
    )


def _is_non_empty_array(value):
    return isinstance(value, list) and value


def _is_non_empty_object(value):
    return isinstance(value, dict) and value



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
#
# Code generated by aaz-dev-tools
# --------------------------------------------------------------------------------------------

# pylint: skip-file
# flake8: noqa

from azure.cli.core.aaz import *


@register_command(
    "netappfiles account backup-vault backup restore-file",
)
class RestoreFile(AAZCommand):
    """Restore the specified files from the specified backup to the active filesystem
    """

    _aaz_info = {
        "version": "2025-01-01",
        "resources": [
            ["mgmt-plane", "/subscriptions/{}/resourcegroups/{}/providers/microsoft.netapp/netappaccounts/{}/backupvaults/{}/backups/{}/restorefiles", "2025-01-01"],
        ]
    }

    AZ_SUPPORT_NO_WAIT = True

    def _handler(self, command_args):
        super()._handler(command_args)
        return self.build_lro_poller(self._execute_operations, None)

    _args_schema = None

    @classmethod
    def _build_arguments_schema(cls, *args, **kwargs):
        if cls._args_schema is not None:
            return cls._args_schema
        cls._args_schema = super()._build_arguments_schema(*args, **kwargs)

        # define Arg Group ""

        _args_schema = cls._args_schema
        _args_schema.account_name = AAZStrArg(
            options=["-a", "--account-name"],
            help="The name of the NetApp account",
            required=True,
            id_part="name",
            fmt=AAZStrArgFormat(
                pattern="^[a-zA-Z0-9][a-zA-Z0-9\\-_]{0,127}$",
            ),
        )
        _args_schema.backup_name = AAZStrArg(
            options=["-b", "--backup-name"],
            help="The name of the backup",
            required=True,
            id_part="child_name_2",
            fmt=AAZStrArgFormat(
                pattern="^[a-zA-Z0-9][a-zA-Z0-9\\-_.]{0,255}$",
            ),
        )
        _args_schema.backup_vault_name = AAZStrArg(
            options=["-v", "--backup-vault-name"],
            help="The name of the Backup Vault",
            required=True,
            id_part="child_name_1",
            fmt=AAZStrArgFormat(
                pattern="^[a-zA-Z0-9][a-zA-Z0-9\\-_]{0,63}$",
            ),
        )
        _args_schema.resource_group = AAZResourceGroupNameArg(
            required=True,
        )

        # define Arg Group "Body"

        _args_schema = cls._args_schema
        _args_schema.destination_volume_id = AAZStrArg(
            options=["--destination-volume-id"],
            arg_group="Body",
            help="Resource Id of the destination volume on which the files need to be restored",
            required=True,
        )
        _args_schema.file_list = AAZListArg(
            options=["--file-list"],
            arg_group="Body",
            help="List of files to be restored",
            required=True,
        )
        _args_schema.restore_file_path = AAZStrArg(
            options=["--restore-file-path"],
            arg_group="Body",
            help="Destination folder where the files will be restored. The path name should start with a forward slash. If it is omitted from request then restore is done at the root folder of the destination volume by default",
            fmt=AAZStrArgFormat(
                pattern="^\\/.*$",
            ),
        )

        file_list = cls._args_schema.file_list
        file_list.Element = AAZStrArg(
            fmt=AAZStrArgFormat(
                max_length=1024,
                min_length=1,
            ),
        )
        return cls._args_schema

    def _execute_operations(self):
        self.pre_operations()
        yield self.BackupsUnderBackupVaultRestoreFiles(ctx=self.ctx)()
        self.post_operations()

    @register_callback
    def pre_operations(self):
        pass

    @register_callback
    def post_operations(self):
        pass

    class BackupsUnderBackupVaultRestoreFiles(AAZHttpOperation):
        CLIENT_TYPE = "MgmtClient"

        def __call__(self, *args, **kwargs):
            request = self.make_request()
            session = self.client.send_request(request=request, stream=False, **kwargs)
            if session.http_response.status_code in [202]:
                return self.client.build_lro_polling(
                    self.ctx.args.no_wait,
                    session,
                    None,
                    self.on_error,
                    lro_options={"final-state-via": "location"},
                    path_format_arguments=self.url_parameters,
                )

            return self.on_error(session.http_response)

        @property
        def url(self):
            return self.client.format_url(
                "/subscriptions/{subscriptionId}/resourceGroups/{resourceGroupName}/providers/Microsoft.NetApp/netAppAccounts/{accountName}/backupVaults/{backupVaultName}/backups/{backupName}/restoreFiles",
                **self.url_parameters
            )

        @property
        def method(self):
            return "POST"

        @property
        def error_format(self):
            return "MgmtErrorFormat"

        @property
        def url_parameters(self):
            parameters = {
                **self.serialize_url_param(
                    "accountName", self.ctx.args.account_name,
                    required=True,
                ),
                **self.serialize_url_param(
                    "backupName", self.ctx.args.backup_name,
                    required=True,
                ),
                **self.serialize_url_param(
                    "backupVaultName", self.ctx.args.backup_vault_name,
                    required=True,
                ),
                **self.serialize_url_param(
                    "resourceGroupName", self.ctx.args.resource_group,
                    required=True,
                ),
                **self.serialize_url_param(
                    "subscriptionId", self.ctx.subscription_id,
                    required=True,
                ),
            }
            return parameters

        @property
        def query_parameters(self):
            parameters = {
                **self.serialize_query_param(
                    "api-version", "2025-01-01",
                    required=True,
                ),
            }
            return parameters

        @property
        def header_parameters(self):
            parameters = {
                **self.serialize_header_param(
                    "Content-Type", "application/json",
                ),
            }
            return parameters

        @property
        def content(self):
            _content_value, _builder = self.new_content_builder(
                self.ctx.args,
                typ=AAZObjectType,
                typ_kwargs={"flags": {"required": True, "client_flatten": True}}
            )
            _builder.set_prop("destinationVolumeId", AAZStrType, ".destination_volume_id", typ_kwargs={"flags": {"required": True}})
            _builder.set_prop("fileList", AAZListType, ".file_list", typ_kwargs={"flags": {"required": True}})
            _builder.set_prop("restoreFilePath", AAZStrType, ".restore_file_path")

            file_list = _builder.get(".fileList")
            if file_list is not None:
                file_list.set_elements(AAZStrType, ".")

            return self.serialize_content(_content_value)


class _RestoreFileHelper:
    """Helper class for RestoreFile"""


__all__ = ["RestoreFile"]
